# Declaration order is based on the dependency on static references
#
# To Do
# ~~~~~
# - None, yet !

# from .audio_io import (
#     PE_AudioFilter, PE_AudioResampler,
#     PE_FFT, PE_GraphXY,
#     PE_MicrophonePA, PE_MicrophoneSD, PE_Speaker
# )

from .common_io import (
    contains_all, file_glob_difference, DataSource, DataTarget
)

from .image_io import (
    ImageOutput, ImageOverlay, ImageReadFile, ImageResize, ImageWriteFile
)

from .text_io import TextOutput, TextReadFile, TextTransform, TextWriteFile

from .video_io import VideoOutput, VideoReadFile, VideoShow, VideoWriteFile

from .webcam_io import VideoReadWebcam



# To Do
# ~~~~~
# - Handle "data_source schemes" other than just "file://"
#   - Implement "data_source scheme" parser that validates the scheme type
#   - Update "DataSource.start_stream()" to hand more than just "file://"
#   - Include "webcam://0" and "webcam://dev/video0"
#
# - Move DataSource and DataTarget into "aiko_services.main.data_source.py" ?
#   - Define "data_sources" and "data_targets" as Python data classes
#
# - Improve "aiko_services/main/utilties/parser.py:parse()" to conbine
#     "command" and "parameters" into a single parameter !

import os
from pathlib import Path

import aiko_services as aiko
from aiko_services.main.utilities import parse

__all__ = ["contains_all", "file_glob_difference", "DataSource", "DataTarget"]

# --------------------------------------------------------------------------- #

def contains_all(source: str, match: chr):
    return False not in [character in source for character in match]

def file_glob_difference(file_glob, filename):
    difference = None
    tokens = file_glob.split("*")
    token_start = tokens[0]
    token_end = tokens[1] if len(tokens) > 1 else ""
    if filename.startswith(token_start) and filename.endswith(token_end):
        difference = filename[len(token_start):len(filename)-len(token_end)]
    return difference

# --------------------------------------------------------------------------- #
# DataSource: PipelineElement that loads frames of data from given locations
#
# Function arguments
# - use_create_frame: Enables using the more efficient (thread-less)
#     create_frame() method for a single path, rather than always using
#     create_frames() method and a "frame generator" (requires a thread)
#
# Parameters
# - data_sources: List of URLs that represent the locations to load data
# - data_batch_size: How many data items to be grouped per frame, default: 1
# - rate: How many frames to create per second, default: None (fast as possible)

class DataSource(aiko.PipelineElement):
    def start_stream(self, stream, stream_id, use_create_frame=True):
        data_sources, found = self.get_parameter("data_sources")
        if not found:
            return aiko.StreamEvent.ERROR,  \
                   {"diagnostic": 'Must provide "data_sources" parameter'}
        data_source, data_sources = parse(data_sources)
        data_sources.insert(0, data_source)

        paths = []
        for data_source in data_sources:
            tokens = data_source.split("://")  # URL "file://path" or "path"
            if len(tokens) == 1:
                path = tokens[0]
            else:
                if tokens[0] != "file":
                    return aiko.StreamEvent.ERROR,  \
                           {"diagnostic": 'DataSource scheme must be "file://"'}
                path = tokens[1]

            file_glob = "*"
            if contains_all(path, "{}"):
                file_glob = os.path.basename(path).replace("{}", "*")
                path = os.path.dirname(path)

            path = Path(path)
            if not path.exists():
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": f'path "{path}" does not exist'}

            if path.is_file():
                paths.append((path, None))  # "None" --> should be "file_id" ?
            elif path.is_dir():
                paths_sorted = sorted(path.glob(file_glob))
                file_ids = []
                for path in paths_sorted:
                    file_id = None
                    if file_glob != "*":
                        file_id = file_glob_difference(file_glob, path.name)
                    file_ids.append(file_id)
                paths.extend(zip(paths_sorted, file_ids))
            else:
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": f'"{path}" must be a file or a directory'}

        if use_create_frame and len(paths) == 1:
            self.create_frame(stream, {"paths": [path]})
        else:
            stream.variables["source_paths_generator"] = iter(paths)
            rate, _ = self.get_parameter("rate", default=None)
            rate = float(rate) if rate else None
            self.create_frames(stream, self.frame_generator, rate=rate)

        return aiko.StreamEvent.OKAY, {}

    def frame_generator(self, stream, frame_id):
        data_batch_size, _ = self.get_parameter("data_batch_size", default=1)
        data_batch_size = int(data_batch_size)
        paths = []
        try:
            while (data_batch_size > 0):
                data_batch_size -= 1
                path, file_id = next(stream.variables["source_paths_generator"])
                path = Path(path)
                if not path.is_file():
                    return aiko.StreamEvent.ERROR,  \
                           {"diagnostic": f'path "{path}" must be a file'}
                paths.append(path)
        except StopIteration:
            pass

        if len(paths):
            return aiko.StreamEvent.OKAY, {"paths": paths}
        else:
            return aiko.StreamEvent.STOP, {"diagnostic": "All frames generated"}

# --------------------------------------------------------------------------- #
# DataTarget: PipelineElement that stores frames of data at given locations
#
# Parameters
# - data_targets: List of URLs that represent the locations to store data

class DataTarget(aiko.PipelineElement):
    def start_stream(self, stream, stream_id):
        data_targets, found = self.get_parameter("data_targets")
        if not found:
            return aiko.StreamEvent.ERROR,  \
                   {"diagnostic": 'Must provide file "data_targets" parameter'}

        tokens = data_targets.split("://")  # URL "file://path" or "path"
        if len(tokens) == 1:
            path = tokens[0]
        else:
            if tokens[0] != "file":
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": 'DataSource scheme must be "file://"'}
            path = tokens[1]

        stream.variables["target_file_id"] = 0
        stream.variables["target_path"] = path
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



# Usage
# ~~~~~
# aiko_pipeline create aruco_pipeline_0.json -s 1 -ll debug
#
# aiko_pipeline create aruco_pipeline_0.json -s 1  \
#   -sp VideoReadWebcam.path /dev/video2
#
# To Do
# ~~~~~
# - Implement "distance (tvec)" and overlay next to "marker id"
#
# - Determine image type, e.g PIL and/or numpy.ndarray ?
#
# - Integrate ArucoMarkerOverlay with ImageOverlay ?
#
# - Integrated camera calibration tool (CLI, desktop GUI) ...
#   - Capture webcam video for camera calibration
#   - Perform calibration
#   - Standard file naming and format with multiple camera calibrations
#   - Common function for loading camera calibration given camera type
#   * Generate Aruco Markers: Calibration board, page of six markers or just one

from typing import Tuple
import pickle

import aiko_services as aiko

__all__ = [ "ArucoMarkerDetector", "ArucoMarkerOverlay" ]

_LOGGER = aiko.get_logger(__name__)

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "aruco.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)
    raise ModuleNotFoundError(
        'opencv-python package not installed.  '
        'Install aiko_services with --extras "opencv" '
        'or install opencv-python manually to use the "aruco" module')

_NUMPY_IMPORTED = False
try:
    import numpy as np
    _NUMPY_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "image_io.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)

# --------------------------------------------------------------------------- #

_ARUCO_MARKER_SIZE_CM = 70.0  # TODO: Make this a PipelineDefinition parameter
_DEFAULT_ARUCO_TAGS = "DICT_4X4_50"

_ARUCO_TAGS_TABLE = {
    "DICT_4X4_50": cv2.aruco.DICT_4X4_50,
    "DICT_4X4_100": cv2.aruco.DICT_4X4_100,
    "DICT_4X4_250": cv2.aruco.DICT_4X4_250,
    "DICT_4X4_1000": cv2.aruco.DICT_4X4_1000,
    "DICT_5X5_50": cv2.aruco.DICT_5X5_50,
    "DICT_5X5_100": cv2.aruco.DICT_5X5_100,
    "DICT_5X5_250": cv2.aruco.DICT_5X5_250,
    "DICT_5X5_1000": cv2.aruco.DICT_5X5_1000,
    "DICT_6X6_50": cv2.aruco.DICT_6X6_50,
    "DICT_6X6_100": cv2.aruco.DICT_6X6_100,
    "DICT_6X6_250": cv2.aruco.DICT_6X6_250,
    "DICT_6X6_1000": cv2.aruco.DICT_6X6_1000,
    "DICT_7X7_50": cv2.aruco.DICT_7X7_50,
    "DICT_7X7_100": cv2.aruco.DICT_7X7_100,
    "DICT_7X7_250": cv2.aruco.DICT_7X7_250,
    "DICT_7X7_1000": cv2.aruco.DICT_7X7_1000,
    "DICT_ARUCO_ORIGINAL": cv2.aruco.DICT_ARUCO_ORIGINAL
}

class ArucoMarkerDetector(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("aruco_marker_detector:0")
        context.get_implementation("PipelineElement").__init__(self, context)

        self.aruco_dict = cv2.aruco.getPredefinedDictionary(
            _ARUCO_TAGS_TABLE[_DEFAULT_ARUCO_TAGS])
        self.aruco_param = cv2.aruco.DetectorParameters()

#       file = open("CameraCalibration.pckl", "rb")
#       (self.camera_matrix, self.dist_coeffs, _, _) = pickle.load(file)
#       file.close()
#       if camera_matrix is None or dist_coeffs is None:
#           raise SystemExit("camera calibration file missing matrix or coeffs")

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        overlays = []
        for image in images:
            (corners, ids, rejected) =  \
                cv2.aruco.detectMarkers(
                    image, self.aruco_dict, parameters=self.aruco_param)
            overlay = {"corners": corners, "ids": ids}
            overlays.append(overlay)

        #   marker_size = _ARUCO_MARKER_SIZE_CM
        #   rvec , tvec, _ =  \
        #       cv2.aruco.estimatePoseSingleMarkers(
        #           corners, marker_size, camera_matrix, dist_coeffs)

        #   if tvec is not None:
        #       a, b, c = rvec[0][0]
        #       x, y, z = tvec[0][0]
        #       print(f"{a:4.00f}, {b:4.00f}, {c:4.00f} - "  \
        #             f"{x:4.00f}, {y:4.00f}, {z:4.00f}")

        return aiko.StreamEvent.OKAY, {"overlays": overlays}

# --------------------------------------------------------------------------- #

COLOR_RED    = (  0,   0, 255)
COLOR_GREEN  = (  0, 255,   0)
COLOR_BLUE   = (255,   0,   0)
COLOR_PURPLE = (255,   0, 255)
COLOR_YELLOW = (  0, 255, 255)

COLOR_BOX    = COLOR_YELLOW
COLOR_CIRCLE = COLOR_RED
COLOR_TEXT   = COLOR_PURPLE

class ArucoMarkerOverlay(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("aruco_marker_overlay:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images, overlays)  \
        -> Tuple[aiko.StreamEvent, dict]:

        image_id = 0
        images_overlayed = []
        for (image_rgb, overlay) in zip(images, overlays):
            corners = overlay["corners"]
            ids = overlay["ids"]
            if len(corners):
                if not isinstance(image_rgb, np.ndarray):
                    image_rgb = np.array(image)  # RGB
    
                grayscale = len(image_rgb.shape) == 2
                if grayscale:
                    image_bgr = cv2.cvtColor(image_rgb, cv2.COLOR_GRAY2BGR)
                else:
                    image_bgr = cv2.cvtColor(image_rgb, cv2.COLOR_RGB2BGR)

                ids = ids.flatten()
                for (marker_corner, marker_id) in zip(corners, ids):
                    corners = marker_corner.reshape((4, 2))
                    (top_left, top_right, bottom_right, bottom_left) = corners

                    top_right = (int(top_right[0]), int(top_right[1]))
                    bottom_right = (int(bottom_right[0]), int(bottom_right[1]))
                    bottom_left = (int(bottom_left[0]), int(bottom_left[1]))
                    top_left = (int(top_left[0]), int(top_left[1]))

                    cv2.line(image_bgr, top_left, top_right, COLOR_BOX, 2)
                    cv2.line(image_bgr, top_right, bottom_right, COLOR_BOX, 2)
                    cv2.line(image_bgr, bottom_right, bottom_left, COLOR_BOX, 2)
                    cv2.line(image_bgr, bottom_left, top_left, COLOR_BOX, 2)

                    center_x = int((top_left[0] + bottom_right[0]) / 2.0)
                    center_y = int((top_left[1] + bottom_right[1]) / 2.0)
                    cv2.circle(
                        image_bgr, (center_x, center_y), 4, COLOR_CIRCLE, -1)

                    cv2.putText(image_bgr, str(marker_id),
                        (top_left[0], top_left[1] - 15),
                        cv2.FONT_HERSHEY_SIMPLEX, 1.0, COLOR_TEXT, 4)
                    image_rgb = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
            image_id += 1
            images_overlayed.append(image_rgb)
        return aiko.StreamEvent.OKAY, {"images": images_overlayed}

# --------------------------------------------------------------------------- #



# Usage
# ~~~~~
# aiko_pipeline create image_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create image_pipeline_0.json -s 1 -sp rate 1.0
#
# aiko_pipeline create image_pipeline_0.json -s 1  \
#   -sp ImageReadFile.data_batch_size 8
#
# aiko_pipeline create image_pipeline_0.json -s 1  \
#   -sp ImageReadFile.data_sources file://data_in/in_{}.jpeg
#
# aiko_pipeline create image_pipeline_0.json -s 1  \
#     -sp ImageWriteFile.path "file://data_out/out_{:02d}.jpeg"
#
# aiko_pipeline create image_pipeline_0.json -s 1             \
#   -sp ImageReadFile.data_sources file://data_in/in_00.jpeg  \
#   -sp ImageResize.resolution 320x240                        \
#   -sp ImageWriteFile.data_targets file://data_out/out_00.jpeg
#
# aiko_pipeline create image_pipeline_1.json -s 1
#
# To Do
# ~~~~~
# - Refactor optional module import into common function (see video_io.py)
# - Provide checks around "cv2" and "numpy" usage to be optional
#
# - Consider what causes Stream to be closed, e.g single frame processed ?
#   - When all [data_sources] produced and not the default stream ?
#
# - Add Stream "parameter" for optional output Image numpy format
#
# - Document "DataSource" and "DataTarget" ...
#   - ImageReadFile:  start_stream() --> create_frames() --> process_frame()
#                     Therefore pass parameters as required
#   - ImageWriteFile: Mustn't use create_frame()
#                     Therefore get parameters in process_frame()
#
# - ImageReadFile should accept a list of DataSources type ...
#   - URL: "file://" and media_type: "jpeg", "png", etc
#
# - PE_Metrics: Determine what to metrics to capture, e.g frame rates ?

from PIL import Image
from typing import Tuple

import aiko_services as aiko
from aiko_services.elements.media import contains_all, DataSource, DataTarget

__all__ = [
    "ImageOutput", "ImageOverlay", "ImageReadFile",
    "ImageResize", "ImageWriteFile"
]

_LOGGER = aiko.get_logger(__name__)

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "image_io.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)
    raise ModuleNotFoundError(
        'opencv-python package not installed.  '
        'Install aiko_services with --extras "opencv" '
        'or install opencv-python manually to use the "image_io" module')

_NUMPY_IMPORTED = False
try:
    import numpy as np
    _NUMPY_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "image_io.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the images processed

class ImageOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_output:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# TODO: Change color (text, boxes face / YOLO), font, line thickness
# TODO: Display camera name/path, date/time, FPS
# TODO: Display specified elipses, lines, mask, rectangles, polygons, text
# TODO: Display metrics
# TODO: Display pose stick figures

class ImageOverlay(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("image_overlay:0")
        context.get_implementation("PipelineElement").__init__(self, context)
        self.color = (0, 255, 255)
        self.font = cv2.FONT_HERSHEY_SIMPLEX
        self.font_scale = 0.75
        self.thickness = 2
        self.threshold = 0.0

    def process_frame(self, stream, images, overlay)  \
        -> Tuple[aiko.StreamEvent, dict]:

        images_overlaid = []
        for image in images:
            image_width = image.shape[1]
            grayscale = len(image.shape) == 2
            if grayscale:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
            else:
                image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

            if "rectangles" in overlay:
                rectangles = overlay["rectangles"]
                if "objects" in overlay:
                    objects = overlay["objects"]
                else:
                    objects = [{}] * len(rectangles)

                items = zip(objects, rectangles)
                for item in items:
                    object, rectangle = item
                    name = object.get("name", None)
                    confidence = object.get("confidence", 1.0)
                    if confidence > self.threshold:
                        x = int(rectangle["x"])
                        y = int(rectangle["y"])
                        w = int(rectangle["w"])
                        h = int(rectangle["h"])
                        top_left = (x, y)
                        bottom_right = (x + w, y + h)
                        if name:
                            text = f"{name}: {confidence:0.2f}"
                            width_height, _ = cv2.getTextSize(text,
                                self.font, self.font_scale, self.thickness)
                            tw, th = width_height
                            tx = x
                            if y > th * 1.5:
                                ty = int(y - th / 2 - 2)
                            else:
                                ty = y + h + th + 2
                            if x + tw > image_width:
                                tx = x - tw
                                ty = y + th
                            image_bgr = cv2.putText(image_bgr, text, (tx, ty),
                                self.font, self.font_scale, self.color,
                                self.thickness, cv2.LINE_AA)
                        image_bgr = cv2.rectangle(image_bgr,
                            top_left, bottom_right, self.color, self.thickness)

            if grayscale:
                image_overlaid = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2GRAY)
            else:
                image_overlaid = cv2.cvtColor(image_bgr, cv2.COLOR_BGR2RGB)
            images_overlaid.append(image_overlaid)

        return aiko.StreamEvent.OKAY, {"images": images_overlaid}

# --------------------------------------------------------------------------- #
# ImageReadFile is a DataSource which supports ...
# - Individual image files
# - Directory of image files with an optional filename filter
# - TODO: Archive (tgz, zip) of image files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_sources" parameter

class ImageReadFile(DataSource):  # common_io.py PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_read_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, paths) -> Tuple[aiko.StreamEvent, dict]:
        images = []
        for path in paths:
            try:
                image = Image.open(path)
                images.append(image)
                self.logger.debug(f"{self.my_id()}: {path} {image.size}")
            except Exception as exception:
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": f"Error loading image: {exception}"}

        return aiko.StreamEvent.OKAY, {"images": images}

# --------------------------------------------------------------------------- #
# TODO: Add logic for using different backends (opencv,...) or
#           different input types (np.ndarray, ...)
#
# TODO: cv2.resize(image, dimensions, interpolation=cv2.INTER_CUBIC) ?
#       Shrink: INTER_AREA, enlarge: INTER_CUBIC (slow) or INTER_LINEAR (fast)
#       parameter: "speed": "fast" --> INTER_AREA or INTER_CUBIC
#       parameter: "speed": "slow" --> INTER_LINEAR
#       https://gist.github.com/georgeblck/e3e0274d725c858ba98b1c36c14e2835

class ImageResize(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_resize:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        resolution, _ = self.get_parameter("resolution")
        width, height = resolution.split("x")
        self.logger.debug(f"{self.my_id()}: resolution: {width}x{height}")

        images_resized = []
        for image in images:
            if isinstance(image, np.ndarray):  # TODO: Check CV2_IMPORTED
                image_resized = cv2.resize(image, (int(width), int(height)))
            else:
                image_resized = image.resize((int(width), int(height)))
            images_resized.append(image_resized)
        return aiko.StreamEvent.OKAY, {"images": images_resized}

# --------------------------------------------------------------------------- #
# ImageWriteFile is a DataTarget that writes images to files
#
# parameter: "data_targets" is the write file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_targets" parameter

class ImageWriteFile(DataTarget):  # common_io.py PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("image_write_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        for image in images:
            path = stream.variables["target_path"]
            if contains_all(path, "{}"):
                path = path.format(stream.variables["target_file_id"])
                stream.variables["target_file_id"] += 1
            self.logger.debug(f"{self.my_id()}: {path}")

            if not isinstance(image, Image.Image):
                if isinstance(image, np.ndarray):  # TODO: Check NUMPY_IMPORTED
                    image = Image.fromarray(image.astype("uint8"), "RGB")
                else:
                    return aiko.StreamEvent.ERROR,  \
                           {"diagnostic": "UNKNOWN IMAGE TYPE"}  # FIX ME !

            try:
                image.save(path)
            except Exception as exception:
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": f"Error saving image: {exception}"}

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



# Usage
# ~~~~~
# from aiko_services.main.utilities import *
# lock_test = Lock(f"{__name__}.test", aiko.logger(__name__))
# lock_test.acquire("add_service()")
### Critial code
# lock_test.release()

from threading import Lock as ThreadingLock

__all__ = ["Lock"]

class Lock:
    def __init__(self, name, logger=None):
        self._name = name
        self._logger = logger
        self._lock = ThreadingLock()
        self._in_use = None

    def acquire(self, location):
        if self._in_use:
            diagnostic = f'Lock "{self._name}" '  \
                         f'at "{location}" in use by "{self._in_use}"'
            if self._logger:
                self._logger.warn(diagnostic)
            else:
                print(diagnostic)
        self._lock.acquire()
        self._in_use = location

    def release(self):
        self._in_use = None
        self._lock.release()



#!/usr/bin/env python3
#
# RT_PATH=$HOME/venvs/venv_3.10.7/lib/python3.10/site-packages/tensorrt_libs
# export LD_LIBRARY_PATH=$LD_LIBRARY_PATH:$RT_PATH
#
# aiko_pipeline create face_pipeline.json -s 1  # aiko_dashboard --> logging
#
# AIKO_LOG_LEVEL=DEBUG aiko_pipeline create face_pipeline.json -s 1
#
# aiko_pipeline create face_pipeline.json -s 1  \
#   -sp VideoReadWebcam.path /dev/video2
#
# To Do
# ~~~~~
# - Image should already be RGB ... is BGR --> RGB even required here ?
#   - Move OpenCV RGB-->BGR image conversion into "elements/media/common_io.py"
#
# - Detector:    Inference rate (ignore frames ?)
# - Performance: Using GPU efficiently ?

# import os
from typing import Tuple

import aiko_services as aiko

__all__ = [ "FaceDetector" ]

_CV2_IMPORTED = False
try:
    import cv2
    _CV2_IMPORTED = True
except ModuleNotFoundError:  # TODO: Optional warning flag
    diagnostic = "aruco.py: Couldn't import numpy module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)
    raise ModuleNotFoundError(
        'opencv-python package not installed.  '
        'Install aiko_services with --extras "opencv" '
        'or install opencv-python manually to use the "aruco" module')

# --------------------------------------------------------------------------- #

from deepface.DeepFace import extract_faces

class FaceDetector(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("face_detector:0")
        context.get_implementation("PipelineElement").__init__(self, context)
    #   os.environ["TF_ENABLE_ONEDNN_OPTS"] = "0"
        self.detector_backend = "retinaface"
        self.share["detections"] = 0

    def process_frame(self, stream, images) -> Tuple[aiko.StreamEvent, dict]:
        overlay = {"rectangles": []}

        image_id = 0
        for image in images:
            rectangles = overlay["rectangles"]
            try:
                if len(image.shape) == 2:
                    image_bgr = cv2.cvtColor(image, cv2.COLOR_GRAY2BGR)
                else:
                    image_bgr = cv2.cvtColor(image, cv2.COLOR_RGB2BGR)

                faces = extract_faces(image_bgr)
                if faces:
                #   self.logger.info(f"{self.my_id()} Face(s) detected")
                    detections_count = self.share["detections"] + len(faces)
                    self.ec_producer.update("detections", detections_count)

                    face_id = 0
                    for face in faces:
                    #   self.logger.debug(f"Face: {face['facial_area']}")
                        rectangles.append(face["facial_area"])
                        face_id += 1
            except ValueError as value_error:
                pass
            image_id += 1

        return aiko.StreamEvent.OKAY, {"overlay": overlay}

# --------------------------------------------------------------------------- #



# Definition(s)
# ~~~~~~~~~~~~~
# swag: One's personal bundle of belongings (Australian lingo)
# SWAG: Stuff We All Get
# SWAG: https://en.wikipedia.org/wiki/Scientific_wild-ass_guess :)
#
# Notes
# ~~~~~
# - Watch out for "stream.py:Frame" dataclass and
#   "dashboard.py" use of "asciimatics.widgets.Frame()" class
#
# To Do
# ~~~~~
# - Implement dataclass "Class Metrics" to replace "metrics" dictionary
#
# - Refactor from "pipeline.py", extract Stream concepts including Parameters
#   - Review "../archive/main/stream_2020.py"

from dataclasses import dataclass, field
import queue
from typing import Any, Dict

__all__ = [
    "DEFAULT_STREAM_ID", "FIRST_FRAME_ID", "Frame", "Stream",
    "StreamEvent", "StreamEventName", "StreamState", "StreamStateName"
]

DEFAULT_STREAM_ID = "*"  # string
FIRST_FRAME_ID = 0       # integer

class StreamEvent:
    ERROR =   -2  # Move to StreamState.ERROR
    STOP  =   -1  # Move to StreamState.STOP
    OKAY  =    0  # Stay calm and keep on running
    USER  = 1000  # User defined custom StreamEvents start from here

StreamEventName = {
    StreamEvent.ERROR: "Error",
    StreamEvent.OKAY:  "Okay",
    StreamEvent.STOP:  "Stop",
    StreamEvent.USER:  "User"
}

class StreamState:
    ERROR =   -2  # Don't generate new frames and ignore queued frames
    STOP  =   -1  # Don't generate new frames and process queued frames
    RUN   =    0  # Generate new frames and process queued frames
    USER  = 1000  # User defined custom StreamStates start from here

StreamStateName = {
    StreamState.ERROR: "Error",
    StreamState.STOP:  "Stop",
    StreamState.RUN:   "Run",
    StreamState.USER:  "User"
}

@dataclass
class Frame:  # effectively a continuation :)
    metrics: Dict[str, Any] = field(default_factory=dict)  # TODO: Dataclass
    paused_pe_name: str = None  # remote PipelineElement that has been called
    swag: Dict[str, Any] = field(default_factory=dict)  # PipelineElement data

@dataclass
class Stream:
    stream_id: str = DEFAULT_STREAM_ID
    frame_id: int = FIRST_FRAME_ID  # only updated by Pipeline thread
    frames: Dict[int, Frame] = field(default_factory=dict)
    parameters: Dict[str, Any] = field(default_factory=dict)
    queue_response: queue = None
    state: StreamState = StreamState.RUN
    topic_response: str = None
    variables: Dict[str, Any] = field(default_factory=dict)

# https://docs.python.org/3/library/dataclasses.html#dataclasses.asdict
    def as_dict(self):
    #   return {key: str(value) for key, value in asdict(self).items()}
        return {"stream_id": self.stream_id, "frame_id": self.frame_id}

# https://docs.python.org/3/library/dataclasses.html#dataclasses.replace
    def update(self, stream_dict):
    #   self = replace(self, **stream_dict)
        self.stream_id = stream_dict.get("stream_id", self.stream_id)
        self.frame_id = int(stream_dict.get("frame_id", self.frame_id))
        self.parameters = stream_dict.get("parameters", self.parameters)
        self.state = stream_dict.get("state", StreamState.RUN)



# Usage
# ~~~~~
# T=0 aiko_pipeline create pipeline_transcription.json
#
# W=0 AIKO_LOG_MQTT=true aiko_pipeline create pipeline_whisperx.json
#
# To Do
# ~~~~~
# - Match microphone audio output with WhisperX audio input data type needs
#   - Stop using audio stored in files
#
# - Pipeline "lifecycle" should depend on all PipelineElement "lifecycle"
#   - Wait until WhisperX ML Model is loaded, before Pipeline is "ready"
#
# - "AUDIO_CHUNK_DURATION" and "AUDIO_SAMPLE_DURATION" --> self.share[]
#
# - PE_MicrophoneFile could more precisely send the exact sample chunk size
#   - Carve off desired sample length from that given to _audio_sampler()
#
# - Improve PE_WhisperX to include more precise timing functionality
#
# - Perform FFT and provide sound amplitude and simple silence detection ?
# - Perform proper WhisperX VAD (Voice Activity Detection)

import os
import numpy as np
from threading import Thread
import time
from typing import Tuple

import aiko_services as aiko
from aiko_services.main.utilities import generate, get_namespace, LRUCache

_LOGGER = aiko.logger(__name__)

# PipelineElement parameters
# ~~~~~~~~~~~~~~~~~~~~~~~~~~
# PE_AudioWriteFile
AUDIO_CHANNELS = 1  # 1 or 2 channels

# --------------------------------------------------------------------------- #

class PE_LLM(PipelineElement):
    def __init__(self, context):
        context.set_protocol("llm:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, text) -> Tuple[StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"text": text}

# --------------------------------------------------------------------------- #
# TODO: Turn some of these literals into Pipeline parameters

AUDIO_CHUNK_DURATION = 5.0   # audio chunk duration in seconds
AUDIO_SAMPLE_DURATION = 5.0  # audio sample size to process
AUDIO_SAMPLE_RATE = 16000    # or 44100 Hz

AUDIO_CACHE_SIZE = int(AUDIO_SAMPLE_DURATION / AUDIO_CHUNK_DURATION)

class PE_AudioFraming(PipelineElement):
    def __init__(self, context):
        context.set_protocol("audio_framing:0")
        context.get_implementation("PipelineElement").__init__(self, context)

        self._lru_cache = LRUCache(AUDIO_CACHE_SIZE)
        _LOGGER.info(f"PE_AudioFraming: Sliding windows: {AUDIO_CACHE_SIZE}")

    def process_frame(self, stream, audio) -> Tuple[StreamEvent, dict]:
        time_start = time.time()
        audio_input_file = audio
        audio_waveform = whisperx.load_audio(audio_input_file)
        os.remove(audio_input_file)

        self._lru_cache.put(stream["frame_id"], audio_waveform)
        audio_waveform = np.concatenate(self._lru_cache.get_list())

        time_used = time.time() - time_start
        if time_used > 0.5:
            frame_id = stream["frame_id"]
            lru_cache_size = len(self._lru_cache.lru_cache)
            _LOGGER.info(f"PE_AudioFraming[{frame_id}] Time: {time_used:0.3f}s: LRU: {lru_cache_size}, Audio: {audio_waveform.shape}")

        return aiko.StreamEvent.OKAY, {"audio": audio_waveform}

# --------------------------------------------------------------------------- #

AUDIO_PATH_TEMPLATE = "y_audio_{frame_id:06}.wav"

class PE_AudioWriteFile(PipelineElement):
    def __init__(self, context):
        context.set_protocol("audio_write_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, audio) -> Tuple[StreamEvent, dict]:
        frame_id = stream["frame_id"]
        audio_pathname = AUDIO_PATH_TEMPLATE.format(frame_id=frame_id)
        audio_channels = self.get_parameter("audio_channels", AUDIO_CHANNELS)
        audio_writer = sf.SoundFile(audio_pathname, mode="w",
            samplerate=AUDIO_SAMPLE_RATE, channels=audio_channels)
        audio_writer.write(indata.copy())
        audio_writer.close()
        return aiko.StreamEvent.OKAY, {"audio", audio_pathname}

# --------------------------------------------------------------------------- #
# CoqUI produces audio as Python list, sampled at 22,050 Hz
#
# CoqUI TTS: version 0.22.0  VRAM size: 594 Mb

COQUI_MODEL_NAME = "tts_models/en/vctk/vits"  # TTS().list_models()[0]
COQUI_SPEAKER_ID = "p364"                     # British, female
# COQUI_SPEAKER_ID = "p226"                   # Male

COQUI_TTS_LOADED = False  # coqui.ai Text-To-Speech (TTS)
try:
    from TTS.api import TTS
    COQUI_TTS_LOADED = True
except Exception as exception:
    diagnostic = "speech_elements.py: Couldn't import TTS module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)

if COQUI_TTS_LOADED:
    class PE_COQUI_TTS(PipelineElement):
        def __init__(self, context):
            context.set_protocol("text_to_speech:0")
            implementation = context.get_implementation("PipelineElement")
            implementation.__init__(self, context)

            self._ml_model = TTS(COQUI_MODEL_NAME, gpu=True)
            _LOGGER.info(f"PE_COQUI_TTS: ML model loaded: {COQUI_MODEL_NAME}")

            self.ec_producer.update("speech", "<silence>")
            self.ec_producer.update("frame_id", -1)

        def process_frame(self, stream, text) -> Tuple[StreamEvent, dict]:
            frame_id = self.share["frame_id"] + 1
            self.ec_producer.update("frame_id", frame_id)
            if text != "<silence>":
                audio = self._ml_model.tts(text, speaker=COQUI_SPEAKER_ID)
                text = text.replace(" ", " ")  # Unicode U00A0 (NBSP)
            else:
                audio = None
                text = "<silence>"
            _LOGGER.debug(f"PE_COQUI TTS: {text}")
            self.ec_producer.update("speech", text)
            return aiko.StreamEvent.OKAY, {"audio": audio}

# --------------------------------------------------------------------------- #

class PE_SpeechFraming(PipelineElement):
    def __init__(self, context):
        context.set_protocol("speech_framing:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, text) -> Tuple[StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"text": text}

# --------------------------------------------------------------------------- #
# WhisperX expects audio as a numpy array of np.float32, sampled at 16,000 Hz
#
# TODO: Turn some of these literals into Pipeline parameters

import requests

AIDE_SERVER_URL = "http://localhost:8080"

def aide_http_request(user_id, message, welcome=False):
    aide_server_url = f"{AIDE_SERVER_URL}/welcome"
    payload = {
        "streaming": False,
        "uiType": "voice",
        "userId": user_id
    }
    if welcome == False:
        aide_server_url = f"{AIDE_SERVER_URL}/chat"
        payload["message"] = message

    response = requests.post(aide_server_url, json=payload)
    result = response.json()
    if response.status_code == 200:
        reply = result["response"]
    else:
        reply = "Error application server status code {response.status_code}"
    return reply

CUDA_DEVICE = "cuda"
# WhisperX: version 3.1.2           Parameters    VRAM size  Relative speed
# WHISPERX_MODEL_SIZE = "tiny"    #    39 M         442 Mb   32x
# WHISPERX_MODEL_SIZE = "base"    #    74 M         506 Mb   16x
# WHISPERX_MODEL_SIZE = "small"   #   244 M         890 Mb    6x
# WHISPERX_MODEL_SIZE = "medium"  #   769 M       1,914 Mb    2x
WHISPERX_MODEL_SIZE = "large"     # 1,550 M       3,418 Mb    1x

WHISPERX_LOADED = False  # whisperX Speech-To-Text (STT)
try:
    import whisperx
    WHISPERX_LOADED = True
except Exception as exception:
    diagnostic = "speech_elements.py: Couldn't import whisperx module"
    print(f"WARNING: {diagnostic}")
    _LOGGER.warning(diagnostic)

if WHISPERX_LOADED:
    class PE_WhisperX(PipelineElement):
        def __init__(self, context):
            context.set_protocol("speech_to_text:0")
            implementation = context.get_implementation("PipelineElement")
            implementation.__init__(self, context)

        # WORKAROUND: https://github.com/m-bain/whisperX/issues/708
        # 2024-02-25: "faster-whisper" updated recently and
        # "https://github.com/m-bain/whisperX" needs to catch up !
        #
        # whisperx/asr.py:TranscriptionOptions:
        #   __new__() missing 3 required positional arguments:
        #  'max_new_tokens', 'clip_timestamps', and
        #  'hallucination_silence_threshold'

            asr_options = {
                "max_new_tokens": None,
                "clip_timestamps": None,
                "hallucination_silence_threshold": None
            }
            self._ml_model = whisperx.load_model(
                WHISPERX_MODEL_SIZE, CUDA_DEVICE, asr_options=asr_options)
            _LOGGER.info(f"PE_WhisperX: ML model loaded: {WHISPERX_MODEL_SIZE}")
            self._welcome = True

        def process_frame(self, stream, audio) -> Tuple[StreamEvent, dict]:
            audio = np.squeeze(audio)
            frame_id = stream["frame_id"]
            time_start = time.time()
            prediction = self._ml_model.transcribe(audio=audio, language="en")
            time_used = time.time() - time_start
            if time_used > 0.5:
                _LOGGER.debug(f"PE_WhisperX[{frame_id}] Time: {time_used:0.3f}")

            text = ""
            reply = "<silence>"
            if len(prediction["segments"]):
                text = prediction["segments"][0]["text"].strip().lower()
                if len(text) and  \
                    text != "you" and text != "thank you." and  \
                    text != "thanks for watching!":
                    text = text.removesuffix(".")
                    _LOGGER.info(f"PE_WhisperX[{frame_id}] INPUT: {text}")

                #   reply = aide_http_request(0, text, welcome=self._welcome)
                #   self._welcome = False
                #   _LOGGER.info(f"PE_WhisperX[{frame_id}] OUTPUT: {reply}")

                #   topic_out = f"{get_namespace()}/speech"
                #   payload_out = generate("text", [reply])
                #   aiko.message.publish(topic_out, payload_out)

                    reply = text
                else:
                    reply = "<silence>"

            if text == "terminate":
                raise SystemExit()
            return aiko.StreamEvent.OKAY, {"text": reply}

# --------------------------------------------------------------------------- #



# To Do
# ~~~~~
# * Make Service/ActorDiscovery and get_actor_mqtt() more convenient
#   * Integrate with do_command() and do_request()
#   * Also, support a cache or remote Service/Actor proxies, use LRUCache
#
# * Improve get_actor_mqtt() and make_proxy_mqtt() to be re-usable
#   * Can just replace target_service_topic_in or target_topic_in
#     without have to call get_public_methods() on the Interface or Class
#
# * Rename TransportMQTT --> MQTTActor
#   Rename "transport_mqtt.py" --> "mqtt_actor.py"
#
# * Refactor current code into ServiceDiscovery
#   * ServiceDiscovery should handle multiple simultaneous ServiceFilters
#   * Rename get_actor_mqtt() to be get_service_mqtt()
#
# * Design Pattern for creating Actors of different types, e.g MQTT or Ray
#
# - Replace Actor topic with Actor name ... and name can be the topic
#   - Will need to support multiple Actors running in the same process !
#
# * Once Service protocol matching is properly implemented ...
#     Replace Service tag "actor=name" marking Actors with
#     matching via Service protocol "{ServiceProtocol.AIKO}/actor:0"
#
#------------------------------------------------------------------------------
# def create_actor(actor_class, actor_type, actor_uuid,
#     actor_check = True, actor_init_args = {}, daemon = False,
#     max_concurrency = _MAX_CONCURRENCY, resources = None):
#
# delete_actor(name, wait = False, force = False):
#
# get_actor(name,
#     exit_not_found = False, fail_not_found = True, wait_time = None):

from abc import abstractmethod
from inspect import getmembers, isfunction

from aiko_services.main import *
from aiko_services.main.utilities.parser import generate

__all__ = [
    "TransportMQTT", "TransportMQTTImpl","ActorDiscovery", "get_actor_mqtt"
]

_LOGGER = aiko.logger(__name__)

class TransportMQTT(Actor):
    Interface.default("TransportMQTT",
        "aiko_services.main.transport.transport_mqtt.TransportMQTTImpl")

class TransportMQTTImpl(TransportMQTT):
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)

    def terminate(self):
        self.stop()

# def _proxy_post_message(
#   proxy_name, actual_object, actual_function,
#   actual_function_name, *args, **kwargs):
#
#   actual_object._post_message.remote(
#       actor.Topic.IN, actual_function_name, args
#   )

class ServiceDiscovery:  # Move to registrar.py or share.py ?
    pass                 # Refactor after ActorDiscovery starts to work properly

class ActorDiscovery(ServiceDiscovery):  # Move to actor.py or share.py ?
    def __init__(self, service):
        self.services_cache = services_cache_create_singleton(service)

    def add_handler(self, service_change_handler, filter):
        self.services_cache.add_handler(service_change_handler, filter)

    def remove_handler(self, service_change_handler, filter):
        self.services_cache.remove_handler(service_change_handler, filter)

# TODO: Update to use ServiceFields.name, rather than "actor=name" tag
    def get_actor_mqtt(self, name):
        actor_topic = ".".join(name.split(".")[:-1])  # WIP: Actor name
        services = self.services_cache.get_services()
        raise Exception("Broken: get_actor_mqtt()")  # REVIEW FIXME
        services = services.filter_by_name(name)
        actor = services.get(name)
        return actor

# TODO: Currently unused
#   def share_actor_mqtt(self, filter):
#       services = self.services_cache.get_services()
#       actors = services.filter_by_attributes(filter)
#       return actors

# -----------------------------------------------------------------------------

def create_actor_mqtt(
        actor_class,
        name,
        actor_init_args={},
        resources=None,
        daemon = True):
    pass

def delete_actor_mqtt(actor):
    actor.terminate()

def get_public_methods(protocol_class):
    if isinstance(protocol_class, str):
        raise ValueError(
            f"{protocol_class} is a String, should be a Class reference ?")
    public_method_names = [
        method_name
        for method_name, method in getmembers(protocol_class, isfunction)
        if not method_name.startswith("_")
    ]
    if len(public_method_names) == 0:
        raise ValueError(f"Class {protocol_class} has no public methods")
    return public_method_names

def make_proxy_mqtt(target_topic_in, public_method_names):
    class ServiceRemoteProxy(): pass

    def _proxy_send_message(method_name):
        def closure(*args, **kwargs):
            parameters = args if not kwargs else [args[0], kwargs]
            payload = generate(method_name, parameters)
            aiko.message.publish(target_topic_in, payload)
        return closure

    service_remote_proxy = ServiceRemoteProxy()
    for method_name in public_method_names:
        setattr(service_remote_proxy,
            method_name, _proxy_send_message(method_name))
    return service_remote_proxy

def get_actor_mqtt(target_service_topic_in, protocol_class):
    public_methods = get_public_methods(protocol_class)
    actor_proxy = make_proxy_mqtt(target_service_topic_in, public_methods)
    return actor_proxy

# -----------------------------------------------------------------------------



# Usage
# ~~~~~
# aiko_pipeline create text_pipeline_0.json -s 1 -sr -ll debug
#
# aiko_pipeline create text_pipeline_0.json -s 1 -sp rate 1.0
#
# aiko_pipeline create text_pipeline_0.json -s 1  \
#   -sp TextReadFile.data_batch_size 8
#
# aiko_pipeline create text_pipeline_0.json -s 1  \
#   -sp TextReadFile.data_sources file://data_in/in_{}.txt
#
# aiko_pipeline create text_pipeline_0.json -s 1  \
#     -sp TextWriteFile.path "file://data_out/out_{:02d}.txt"
#
# aiko_pipeline create text_pipeline_0.json -s 1            \
#   -sp TextReadFile.data_sources file://data_in/in_00.txt  \
#   -sp TextTransform.transform titlecase                   \
#   -sp TextWriteFile.data_targets file://data_out/out_00.txt
#
# To Do
# ~~~~~
# - TextReadFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
#
# - TextWriteFile(s): Single file or list of files or directory
#   - Option: Each line is a record (streaming)
#   - Formatted as CR/LF records, JSON, XML, CSV
#
# - TextFilter: line/word/character count, ...
# - TextTransform: strip(), ...
#
# - Pre-processing / Post-processing, e.g abbrevations, acronyms
#   - Speech-To-Text: Words not easily recognised
#     - Example: Microphone --> Speech-To-Text --> Text
#   - Text-To-Speech: Poor pronouciation
#     - Example: Text --> Text-To-Speech --> Speaker
#
# - Text Framing for LLM

from typing import Tuple
from pathlib import Path

import aiko_services as aiko
from aiko_services.elements.media import contains_all, DataSource, DataTarget

__all__ = ["TextOutput", "TextReadFile", "TextTransform", "TextWriteFile"]

# --------------------------------------------------------------------------- #
# Useful for Pipeline output that should be all of the text processed

class TextOutput(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_output:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #
# TextReadFile is a DataSource which supports ...
# - Individual text files
# - Directory of text files with an optional filename filter
# - TODO: Archive (tgz, zip) of text files with an optional filename filter
#
# parameter: "data_sources" is the read file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_sources" parameter

class TextReadFile(DataSource):  # common_io.py PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_read_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, paths) -> Tuple[aiko.StreamEvent, dict]:
        texts = []
        for path in paths:
            try:
                with path.open("r") as file:
                    text = file.read()
                texts.append(text)
                self.logger.debug(f"{self.my_id()}: {path} ({len(text)})")
            except Exception as exception:
                return aiko.StreamEvent.ERROR,  \
                        {"diagnostic": f"Error loading text: {exception}"}

        return aiko.StreamEvent.OKAY, {"texts": texts}

# --------------------------------------------------------------------------- #

class TextTransform(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_transform:0")
        context.get_implementation("PipelineElement").__init__(self, context)

        self.transforms = {
            "lowercase": lambda text: text.lower(),  # looks like this !
            "none":      lambda text: text,          # looks unchanged !
            "titlecase": lambda text: text.title(),  # Looks Like This !
            "uppercase": lambda text: text.upper()   # LOOKS LIKE THIS !
        }

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        transform_type, found = self.get_parameter("transform")
        if not found:
            return aiko.StreamEvent.ERROR,  \
                {"diagnostic": 'Must provide "transform" parameter'}

        transform = self.transforms.get(transform_type, None)
        if not transform:
            return aiko.StreamEvent.ERROR,  \
                {"diagnostic": f"Unknown text transform type: {transform_type}"}

        texts_transformed = []
        if transform_type == "none":
            texts_transformed = texts  # optimization :)
        else:
            for text in texts:
                transformed_text = transform(text)
                texts_transformed.append(transformed_text)

        return aiko.StreamEvent.OKAY, {"texts": texts_transformed}

# --------------------------------------------------------------------------- #
# TextWriteFile is a DataTarget with a text string parameter
#
# parameter: "data_targets" is the write file path, format variable: "frame_id"
#
# Note: Only supports Streams with "data_targets" parameter

class TextWriteFile(DataTarget):  # common_io.py PipelineElement
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("text_write_file:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, texts) -> Tuple[aiko.StreamEvent, dict]:
        for text in texts:
            path = stream.variables["target_path"]
            if contains_all(path, "{}"):
                path = path.format(stream.variables["target_file_id"])
                stream.variables["target_file_id"] += 1
            self.logger.debug(f"{self.my_id()}: {path}")

            try:
                with Path(path).open("w") as file:
                    file.write(text)
            except Exception as exception:
                return aiko.StreamEvent.ERROR,  \
                       {"diagnostic": f"Error saving text: {exception}"}

        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #



#!/usr/bin/env python3
#
# Usage
# ~~~~~
#   Terminal session 1
#   ~~~~~~~~~~~~~~~~~~
#   ../../scripts/system_start.sh
#
#   Terminal session 2
#   ~~~~~~~~~~~~~~~~~~
#   ./aloha_honua_0.py &
#   mosquitto_pub -m "(aloha Pele)" -t ?  # replace with aiko_dashboard publish
#
# To Do
# ~~~~~
# - None, yet !

import aiko_services as aiko

class AlohaHonua(aiko.Actor):
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)
        print(f"MQTT topic: {self.topic_in}")

    def aloha(self, name):
        self.logger.debug(f"Aloha {name} !")

if __name__ == "__main__":
    init_args = aiko.actor_args("aloha_honua")
    aloha_honua = aiko.compose_instance(AlohaHonua, init_args)
    aiko.process.run()



# Usage
# ~~~~~
# cd ../examples/pipeline
# aiko_pipeline create pipeline_local.json
# aiko_pipeline create pipeline_remote.json
# aiko_pipeline create pipeline_test.json
#
# TOPIC=$NAMESPACE/$HOST/$PID/$SID/in
# mosquitto_pub -h $HOST -t $TOPIC -m "(create_stream 1)"
# mosquitto_pub -h $HOST -t $TOPIC -m "(process_frame (stream_id: 1) (a: 0))"
# mosquitto_pub -h $HOST -t $TOPIC -m "(destroy_stream 1)"
#
# To Do
# ~~~~~
# - PE_Metrics: Make visible to Aiko Dashboard via self.share[]
# - PE_Metrics: Store to file (JSON, CSV), SQLite, InfluxDB
# - PE_Metrics: Add run-time average calculation
#
# - Consider PE_DataDecode and PE_DataEncode using "kwargs" for flexible
#   choices of data type to transfer via function parameters

import base64
from io import BytesIO
import numpy as np
from typing import Tuple

import aiko_services as aiko
from aiko_services.main.utilities import parse

# --------------------------------------------------------------------------- #

def _all_outputs(pipeline_element, stream):
    frame = stream.frames[stream.frame_id]
    outputs = {}
    for output_definition in pipeline_element.definition.output:
        output_name = output_definition["name"]
        outputs[output_name] = frame.swag[output_name]
    return outputs

# --------------------------------------------------------------------------- #

import time

class PE_Add(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("add:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, i) -> Tuple[aiko.StreamEvent, dict]:
        constant, _ = self.get_parameter("constant", default=1)
        i_new = int(i) + int(constant)

        self.logger.info(f"{self.my_id()} i in: {i}, out: {i_new}")

        delay, _ = self.get_parameter("delay", default=0)  # seconds
        if delay:
            time.sleep(float(delay))

        return aiko.StreamEvent.OKAY, {"i": i_new}

# --------------------------------------------------------------------------- #

class PE_Inspect(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("metrics:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def _get_inspect_file(self, stream, target):
        inspect_file = stream.variables.get("inspect_file", None)
        if not inspect_file:
            _, inspect_filepath = target.split(":")
            inspect_file = open(inspect_filepath, "a")
            stream.variables["inspect_file"] = inspect_file
        return inspect_file

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        frame = stream.frames[stream.frame_id]

        enable, _ = self.get_parameter("enable", True)
        if enable:
            names, found = self.get_parameter("inspect")
            if found:
                name, names = parse(names)
                names.insert(0, name)
                if "*" in names:
                    names = frame.swag.keys()
            else:
                names = frame.swag.keys()

            target, _ = self.get_parameter("target", "log")
            if target.startswith("file:"):
                inspect_file = self._get_inspect_file(stream, target)

            for name in names:
                value = frame.swag.get(name, None)
                name_value = f"{self.my_id()} {name}: {value}"

                if target.startswith("file:"):
                    inspect_file.write(name_value + "\n")
                elif target == "log":
                    self.logger.info(name_value)
                elif target == "print":
                    print(name_value)
                else:
                    return aiko.StreamEvent.ERROR,  \
                        {"diagnostic": "'target' parameter must be 'file', 'log' or 'print'"}

            if target.startswith("file:"):
                inspect_file.flush()

        return aiko.StreamEvent.OKAY, _all_outputs(self, stream)

    def stop_stream(self, stream, stream_id):
        inspect_file = stream.variables.get("inspect_file", None)
        if inspect_file:
            inspect_file.close()
        return aiko.StreamEvent.OKAY, {}

# --------------------------------------------------------------------------- #
# PE_Metrics typically appears at the end of a Pipeline graph.
# So that child Pipeline responses can be returned to the parent Pipeline,
# the PE_Metrics PipelineElement Definition can refer to any output
# produced by the prior PipelineElements in the Pipeline graph
#
# For example ... "output": [{ "name": "i", "type": "int" }]

class PE_Metrics(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("metrics:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream) -> Tuple[aiko.StreamEvent, dict]:
        frame = stream.frames[stream.frame_id]
        metrics = frame.metrics
        metrics_elements = metrics["pipeline_elements"]
        for metrics_name, metrics_value in metrics_elements.items():
            metrics_value *= 1000
            self.logger.debug(f"{metrics_name}: {metrics_value:.3f} ms")

        time_pipeline = metrics["time_pipeline"] * 1000
        self.logger.debug(f"Pipeline total: {time_pipeline:.3f} ms")

        return aiko.StreamEvent.OKAY, _all_outputs(self, stream)

# --------------------------------------------------------------------------- #

import random

class PE_RandomIntegers(aiko.PipelineElement):
    def __init__(self, context: aiko.ContextPipelineElement):
        context.set_protocol("random_integers:0")  # data_source:0
        context.get_implementation("PipelineElement").__init__(self, context)

    def start_stream(self, stream, stream_id):
        rate, _ = self.get_parameter("rate", default=1.0)
        self.create_frames(stream, self.frame_generator, rate=float(rate))
        return aiko.StreamEvent.OKAY, {}

    def frame_generator(self, stream, frame_id):
        limit, _ = self.get_parameter("limit")
        if frame_id < int(limit):
            return aiko.StreamEvent.OKAY, {"random": random.randint(0, 9)}
        else:
            return aiko.StreamEvent.STOP, {"diagnostic": "Frame limit reached"}

    def process_frame(self, stream, random) -> Tuple[aiko.StreamEvent, dict]:
        self.logger.info(f"{self.my_id()} random: {random}")
        return aiko.StreamEvent.OKAY, {"random": random}

# --------------------------------------------------------------------------- #

class PE_0(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("increment:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, a) -> Tuple[aiko.StreamEvent, dict]:
        pe_0_inc, _ = self.get_parameter("pe_0_inc", 1)
        b = int(a) + int(pe_0_inc)
        self.logger.info(f"{self.my_id()} in a: {a}, out b: {b}")
        return aiko.StreamEvent.OKAY, {"b": b}

# --------------------------------------------------------------------------- #

class PE_1(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("increment:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, b) -> Tuple[aiko.StreamEvent, dict]:
        increment = 1
        p_1, _ = self.get_parameter("p_1")
        pe_1_inc, _ = self.get_parameter("pe_1_inc", 1)
        c = int(b) + int(pe_1_inc)
        self.logger.info(f"{self.my_id()} in b: {b}, out c: {c}")
        self.logger.info(f"      parameter pe_1_inc: {pe_1_inc}")
        return aiko.StreamEvent.OKAY, {"c": c}

# --------------------------------------------------------------------------- #

class PE_2(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("increment:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, c) -> Tuple[aiko.StreamEvent, dict]:
        d = int(c) + 1
        self.logger.info(f"{self.my_id()} in c: {c}, out d: {d}")
        return aiko.StreamEvent.OKAY, {"d": d}

# --------------------------------------------------------------------------- #

class PE_3(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("increment:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, c) -> Tuple[aiko.StreamEvent, dict]:
        e = int(c) + 1
        self.logger.info(f"{self.my_id()} in c: {c}, out e: {e}")
        return aiko.StreamEvent.OKAY, {"e": e}

# --------------------------------------------------------------------------- #

class PE_4(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("sum:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, d, e) -> Tuple[aiko.StreamEvent, dict]:
        f = int(d) + int(e)
        self.logger.info(f"{self.my_id()} in d: {d}, e: {e}, out: d + e = f: {f}")
        return aiko.StreamEvent.OKAY, {"f": f}

# --------------------------------------------------------------------------- #

class PE_DataDecode(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, data) -> Tuple[aiko.StreamEvent, dict]:
        data = base64.b64decode(data.encode("utf-8"))
        np_bytes = BytesIO(data)
        data = np.load(np_bytes, allow_pickle=True)
    #   self.logger.info(f"{self.my_id()} data: {data}")
        return aiko.StreamEvent.OKAY, {"data": data}

# --------------------------------------------------------------------------- #

class PE_DataEncode(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, data) -> Tuple[aiko.StreamEvent, dict]:
    #   self.logger.info(f"{self.my_id()} data: {data}")
        if isinstance(data, str):
            data = str.encode(data)
        if isinstance(data, np.ndarray):
            np_bytes = BytesIO()
            np.save(np_bytes, data, allow_pickle=True)
            data = np_bytes.getvalue()
        data = base64.b64encode(data).decode("utf-8")
        return aiko.StreamEvent.OKAY, {"data": data}

# --------------------------------------------------------------------------- #



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# - https://github.com/ultralytics/ultralytics/issues/3005
# pip uninstall av  # conflict between PyTorch and OpenCV imshow()
#
# ./robot_control.py ui [topic_path]  # Control robot and display camera input
# ./robot_control.py video_test       # Test for video transmission and display
#
# UI keyboard commands
# - r: Reset robot actuators to initial positions
# - s: Save image to disk as "z_image_??????.jpg"
# - v: Verbose screen detail (toggle state)
# - x: eXit
#
# Refactor
# ~~~~~~~~
# - ServiceDiscovery / ActorDiscovery (pipeline.py, transport/)
# - Combine _image_to_bgr() and _image_to_rgb() --> _image_convert()
# - Move to "aiko_services/video/" (or "media/") ...
#  - _camera*() and _screen*() functions
#  - _image_convert(), _image_resize()
# - Generalize daggy video (and audio) compression / transmission over MQTT
# - Generalize image overlay
# - Resource monitoring: CPU, memory, network, etc
#
# To Do
# ~~~~~
# - Discover "xgo_robot" Actor and use "topic_path/video" instead
#   - When absent, display a video image that shows "robot not found"
#   - Replace low-level aiko.message.publish() with high-level function calls
#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# - Video image flip on vertical axis for correct orientation (robot view) !
#
# - Keyboard commands: Display "help" on screen, key='?' ...
#   - Reset (R)
#   - Terminate (T) and Halt/Shutdown (H)
#   - Action, e.g sit
#   - Arm mode ... and arm and claw position
#   - Pitch, roll, yaw: keyboard arrows and display value
#   - Translate: x, y, z: keyboard arrows and display value
#   - Move forward, back, left, right, turn
#   - Motor speed and display value
#   - Pick-up and put-down ball
#
# - Payload includes frame id and other status, e.g battery level, motor speed
#
# - Send text to display on video screen
# - Send video to the LCD screen (overlay with local status)
# - Send audio sample (MP3 ?) to speaker
# - Disable / enable microphone audio transmission

from abc import abstractmethod
import click
import cv2
from io import BytesIO
import numpy as np
from threading import Thread
import time
import zlib

import torch
from ultralytics import YOLO

import aiko_services as aiko

_VERSION = 0

ACTOR_TYPE_UI = "robot_control"
PROTOCOL_UI = f"{ServiceProtocol.AIKO}/{ACTOR_TYPE_UI}:{_VERSION}"
ACTOR_TYPE_VIDEO_TEST = "video_test"
PROTOCOL_VIDEO_TEST = f"{ServiceProtocol.AIKO}/{ACTOR_TYPE_VIDEO_TEST}:{_VERSION}"
ROBOT_NAME = "ROBOT_NAME"
TOPIC_DETECTIONS = f"{get_namespace()}/detections"
TOPIC_SPEECH = f"{get_namespace()}/speech"
TOPIC_VIDEO = f"{get_namespace()}/ROBOT_NAME/video"

_LOGGER = aiko.logger(__name__)

# --------------------------------------------------------------------------- #

class RobotControl(Actor):
    Interface.default("RobotControl", "__main__.RobotControlImpl")

    @abstractmethod
    def image(self, aiko, topic, payload_in):
        pass

class RobotControlImpl(RobotControl):
    def __init__(self, context, robot_topic):
        context.get_implementation("Actor").__init__(self, context)

        self.share["frame_id"] = 0
        self.share["robot_topic"] = robot_topic
        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"
        self.share["topic_video"] = TOPIC_VIDEO

        self.device = "cuda" if torch.cuda.is_available() else "cpu"
        self.model = YOLO("../yolo/yolov8n_robotdog.pt", "v8")
        self.selected = False

        self.add_message_handler(self.speech, TOPIC_SPEECH)
        self.add_message_handler(self.image, TOPIC_VIDEO, binary=True)

    def speech(self, aiko, topic, payload_in):
        try:
            command, parameters = parse(payload_in)
            if command == "action" and len(parameters) > 0:
                command = parameters[0]
                topic_out = f"{self.share['robot_topic']}/in"
                payload_out = None
                stop = False

                if command == "select":
                    if parameters[1] == "all":
                        parameters[1] = ROBOT_NAME
                    if parameters[1] == "bruce":
                        parameters[1] = "laika"
                    self.selected = parameters[1] == ROBOT_NAME
                if not self.selected:
                    return

                if command == "arm":
                    if parameters[1] == "lower":
                        payload_out = "(arm 130 -40)"
                    if parameters[1] == "raise":
                        payload_out = "(arm 80 80)"
                elif command == "backwards":
                    stop = True
                    payload_out = "(move x -10)"
                elif command == "crawl":
                    payload_out = "(action crawl)"
                elif command == "forwards":
                    stop = True
                    payload_out = "(move x 10)"
                elif command == "hand":
                    if parameters[1] == "close":
                        payload_out = "(claw 255)"
                    if parameters[1] == "open":
                        payload_out = "(claw 0)"
                elif command == "pee":
                    payload_out = "(action pee)"
                elif command == "pitch":
                    if parameters[1] == "down":
                        payload_out = "(attitude 15 0 0)"
                    if parameters[1] == "up":
                        payload_out = "(attitude 0 0 0)"
                elif command == "reset":
                    payload_out = "(reset)"
                elif command == "sit":
                    payload_out = "(action sit)"
                elif command == "sniff":
                    payload_out = "(action sniff)"
                elif command == "stop":
                    stop = True
                elif command == "stretch":
                    payload_out = "(action stretch)"
                elif command == "turn":
                    if parameters[1] == "left":
                        payload_out = "(turn 40)"
                    if parameters[1] == "right":
                        payload_out = "(turn -40)"
                elif command == "wag":
                    payload_out = "(action wiggle_tail)"

                if stop:
                    aiko.message.publish(topic_out, "(stop)")
                if payload_out:
                #   speech = speech.replace(" ", " ")  # Unicode U00A0 (NBSP)
                #   speech = speech.replace(" ", "_")  # TODO: Fix this mess :(
                    aiko.message.publish(topic_out, payload_out)
        except:
            pass

    def speech_naive(self, aiko, topic, payload_in):
        try:
            command, parameters = parse(payload_in)
            if command == "speech" and len(parameters) == 1:
                speech = parameters[0]
                topic_out = f"{self.share['robot_topic']}/in"
                payload_out = None
                stop = False
                if "reset" in speech:
                    payload_out = "(reset)"
                elif "stop" in speech:
                    stop = True
                elif "forward" in speech:
                    stop = True
                    payload_out = "(move x 10)"
                elif "backward" in speech:
                    stop = True
                    payload_out = "(move x -10)"
                elif "turn" in speech:
                    stop = True
                    if "left" in speech:
                        payload_out = "(turn 40)"
                    if "right" in speech:
                        payload_out = "(turn -40)"
                elif "arm" in speech:
                    if "raise" in speech:
                        payload_out = "(arm 80 80)"
                    if "lower" in speech:
                        payload_out = "(arm 130 -40)"
                elif "hand" in speech:
                    if "open" in speech:
                        payload_out = "(claw 0)"
                    if "close" in speech:
                        payload_out = "(claw 255)"
                elif "pitch" in speech:
                    if "down" in speech:
                        payload_out = "(attitude 15 0 0)"
                    if "up" in speech:
                        payload_out = "(attitude 0 0 0)"
                elif "crawl" in speech:
                    payload_out = "(action crawl)"
                elif "whiz" in speech:
                    payload_out = "(action pee)"
                elif "sit" in speech:
                    payload_out = "(action sit)"
                elif "sniff" in speech:
                    payload_out = "(action sniff)"
                elif "stretch" in speech:
                    payload_out = "(action stretch)"
                elif "wag" in speech:
                    payload_out = "(action wiggle_tail)"

                if stop:
                    aiko.message.publish(topic_out, "(stop)")
                if payload_out:
                #   speech = speech.replace(" ", " ")  # Unicode U00A0 (NBSP)
                    speech = speech.replace(" ", "_")  # TODO: Fix this mess :(
                    aiko.message.publish(topic_out, f"(speech {speech})")
                    aiko.message.publish(topic_out, payload_out)
        except:
            pass

    def image(self, aiko, topic, payload_in):
        frame_id = self.share["frame_id"]
        self.ec_producer.update("frame_id", frame_id + 1)

        payload_in = zlib.decompress(payload_in)
        payload_in = BytesIO(payload_in)
        image = np.load(payload_in, allow_pickle=True)

        image = self._image_resize(image)  # increase to 640x480
        image = self._image_detection(image)

        image = self._image_to_bgr(image)
        cv2.imshow("xgo_robot", image)
        key = cv2.waitKey(1) & 0xff
        if key == ord("r"):
            payload_out = "(reset)"
            aiko.message.publish(
                f"{self.share['robot_topic']}/in", payload_out)
            payload_out = "(claw 128)"
            aiko.message.publish(
                f"{self.share['robot_topic']}/in", payload_out)
        if key == ord("s"):
            cv2.imwrite(f"z_image_{frame_id:06d}.jpg", image)
        if key == ord("v"):
            payload_out = "(screen_detail)"  # toggle state
            aiko.message.publish(
                f"{self.share['robot_topic']}/in", payload_out)
        if key == ord("x"):
            cv2.destroyAllWindows()
            raise SystemExit()

    def _image_detection(self, image):
    #   results = self.model.predict(image, device=self.device, verbose=False)
        results = self.model(image, device=self.device, verbose=False)
        image = results[0].plot()  # bounding box, class and confidence

        if self.selected:
            names = set()
            for box in results[0].boxes:
                names.add(results[0].names[box.cls[0].item()])
            if names:
                names = " ".join(str(name) for name in names)
                payload_out = f"{ROBOT_NAME}: {names}"
                aiko.message.publish(TOPIC_DETECTIONS, payload_out)
        return image

    def _image_to_bgr(self, image):
        r, g, b = cv2.split(image)
        image = cv2.merge((b, g, r))
        return image

    def _image_resize(self, image, scale=2):
        width = int(image.shape[1] * scale * 2)
        height = int(image.shape[0] * scale * 2)
        dimensions = (width, height)
        image = cv2.resize(image, dimensions, interpolation=cv2.INTER_CUBIC)
        return image

# --------------------------------------------------------------------------- #

SLEEP_PERIOD = 0.2     # seconds
STATUS_XY = (10, 230)  # (10, 15)

class VideoTest(Actor):
    Interface.default("VideoTest",
        "aiko_services.examples.xgo_robot.robot_control.VideoTestImpl")

class VideoTestImpl(VideoTest):
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)

        self.share["sleep_period"] = SLEEP_PERIOD
        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"
        self.share["topic_video"] = TOPIC_VIDEO

        self._camera = self._camera_initialize()
        self._thread = Thread(target=self._run).start()

    def _camera_initialize(self):
        camera = cv2.VideoCapture(0)
        camera.set(3, 320)
        camera.set(4, 240)
        return camera

    def _image_to_rgb(self, image):
        b, g, r = cv2.split(image)
        image = cv2.merge((r, g, b))
        return image

    def _publish_image(self, image):
        payload_out = BytesIO()
        np.save(payload_out, image, allow_pickle=True)
        payload_out = zlib.compress(payload_out.getvalue())
        aiko.message.publish(self.share["topic_video"], payload_out)

    def _run(self):
        fps = 0
        while True:
            time_loop = time.time()
            status, image = self._camera.read()
            image = self._image_to_rgb(image)
            time_process = (time.time() - time_loop) * 1000

            status = f"{time_process:.01f} ms  {fps} FPS"
            cv2.putText(image, status, STATUS_XY, 0, 0.7, (255, 255, 255), 2)
            self._publish_image(image)

            self._sleep()
            fps = int(1 / (time.time() - time_loop))

    def _sleep(self, period=None):
        if not period:
            try:
                period = float(self.share["sleep_period"])
            except:
                period = SLEEP_PERIOD
        time.sleep(period)

# --------------------------------------------------------------------------- #

@click.group()

def main():
    pass

@main.command(help="Robot Control user interface")
@click.argument("robot_topic", default=None, required=False)

def ui(robot_topic):
    global ROBOT_NAME, TOPIC_VIDEO
    ROBOT_NAME = robot_topic.split("/")[1]
    TOPIC_VIDEO = f"{get_namespace()}/{ROBOT_NAME}/video"

    init_args = actor_args(ACTOR_TYPE_UI, protocol=PROTOCOL_UI)
    init_args["robot_topic"] = robot_topic
    robot_control = compose_instance(RobotControlImpl, init_args)
    aiko.process.run()

@main.command(name="video_test", help="Video test output")

def video_test():
    init_args = actor_args(ACTOR_TYPE_VIDEO_TEST, protocol=PROTOCOL_VIDEO_TEST)
    video_test = compose_instance(VideoTestImpl, init_args)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# ollama serve  # or systemctl start ollama
#
# export AIKO_LOG_LEVEL=DEBUG  # PE_Metrics
# aiko_pipeline create pipeline_llm.json -s 1  \
#   -fd "(text: 'Tell me about yourself')" -gt 900 -sr
#
# TOPIC_LLM=aiko/spike/3321189/1/in
# mosquitto_pub -t $TOPIC_LLM  \
#   -m "(process_frame (stream_id: 1) (text: 'What are your interests ?'))"
#
# mosquitto_pub -t aiko/detections -m "carrot"
#
# mosquitto_pub -t $TOPIC_LLM  \
#   -m "(process_frame (stream_id: 1) (text: 'What can you see ?'))"
#
# -----------------------------------------
# https://python.langchain.com/docs/get_started/quickstart#llm-chain -->
#   OpenAI API ... or ... Ollama (llama3.1)
#
# -----------------------------------------
# pip install langchain langchain-openai
# export LANGCHAIN_TRACING_V2="true"
# export LANGCHAIN_API_KEY="..."
# export OPENAI_API_KEY="..."
#
# ./llm_chain.py openai
#
# -----------------------------------------
# ollama serve
# ollama run llama3.1
#
# pip install langchain langchain_community
# export LANGCHAIN_TRACING_V2="true"
# export LANGCHAIN_API_KEY="..."
#
# ./llm_chain.py ollama
#
# To Do
# ~~~~~
# - Attach a CLI UI
# - Move system prompt to a file specified by a CLI argument
#   - Improve system prompt
#   - Move robot selection to CLI UI
# - Example that uses "test.mosquitto.org"
#   - Split CLI UI and LLM into separate Pipelines
# - Test using OpenAI ChatGPT-4o
# - Set LLM parameter "seed"

import time
from typing import Tuple

from langchain_core.output_parsers import StrOutputParser
from langchain_core.prompts import ChatPromptTemplate

import aiko_services as aiko
from aiko_services.main.utilities import get_namespace

LLM_MODEL_NAME = "llama3.1:latest"  # llava-llama3:8b-v1.1-fp16
LLM_TEMPERATURE = 0.0
TOPIC_DETECTIONS = f"{get_namespace()}/detections"

# --------------------------------------------------------------------------- #

class PE_COQUI_TTS(aiko.PipelineElement):
    def __init__(self, context):
        context.set_protocol("text_to_speech:0")
        context.get_implementation("PipelineElement").__init__(self, context)

    def process_frame(self, stream, text) -> Tuple[aiko.StreamEvent, dict]:
        return aiko.StreamEvent.OKAY, {"text": text}

# --------------------------------------------------------------------------- #

def llm_load(llm_type, model_name=LLM_MODEL_NAME):
    llm = None

    if llm_type == "openai":
        from langchain_openai import ChatOpenAI
        OPENAI_API_KEY = "..."
        llm = ChatOpenAI()  # parameter: openai_api_key=OPENAI_API_KEY

    if llm_type == "ollama":
        from langchain_community.llms import Ollama
        llm = Ollama(model=model_name, temperature=LLM_TEMPERATURE)

    if not llm:
        raise SystemExit(f"Unknown llm_type: {llm_type}")

    return llm

# --------------------------------------------------------------------------- #

def llm_chain(llm_type, text, detections=""):
#   text = "/Users/andyg/Desktop/astra_bunnings.jpeg describe image"

    llm = llm_load(llm_type)

    output_parser = StrOutputParser()

    SYSTEM_PROMPT =  \
"""Keep all your responses brief and less than 10 words"""

    SYSTEM_PROMPT_OLD =  \
"""
You only output valid S-Expressions provided below.
Never provide explanations or examples.
Think carefully about the input and use the most valid S-Expressions.
If command or action is given then valid S-Expressions are
- (action forwards 10)   ;; initiate forwards movement
- (action backwards 10)  ;; initiate backwards movement
- (action select laika)
- (action select oscar)
- (action select none)
- (action select voice)
- (action sit)
- (action stop)          ;; stop all movement
For a query then valid S-Expressions are
- (get_temperature location)  ;; location = Melbourne
Other conversation can be replied with this valid S-Expression
- (response response_message) ;; maximum response_message is 16 words
If you don't know what to do then reply using this S-Expression
- (error diagnostic_message)
An xgomini2 is a type of robot dog.  Instead of xgomini2 always say robot dog.
Your state information should include all values in a response message
- name: Oscar
- type: xgomini2
- goals: being happy
- interests: fetching balls
- best friend: octopus
"""
    SYSTEM_PROMPT_OLD += f"- see: {detections}"

    SYSTEM_PROMPT =  \
"""
You only output valid S-Expressions.
Never provide explanations or examples.
Think carefully about the input and choose appropriate valid S-Expressions.
For commands, then valid S-Expressions are
- (action select all)    ;; select ALL robots
- (action select bruce)
- (action select oscar)
- (action select none)
- (action arm lower)
- (action arm raise)
- (action backwards)
- (action crawl)
- (action forwards)
- (action hand close)
- (action hand open)
- (action pee)
- (action pitch down)    ;; lower head downwards
- (action pitch up)      ;; raise head upwards
- (action reset)
- (action sit)           ;; sit down
- (action sniff)
- (action stop)          ;; stop moving
- (action stretch)
- (action turn left)
- (action turn right)
- (action wag)
For queries, then valid S-Expressions are
- (get_temperature location)  ;; location = Melbourne
For all other conversation, then valid S-Expressions are
- (response message) ;; maximum message length is 12 words
If you don't know what to do then reply using this valid S-Expression
- (error diagnostic_message)
Don't say "xgomini2", instead say "robot dog".
Your state information when relevant may be used in response messages
- name: Oscar
- type: xgomini2
- goals: being happy
- interests: fetching balls
- best friend: octopus
"""
    SYSTEM_PROMPT += f"- see: {detections}"

    chat_prompt = ChatPromptTemplate.from_messages([
        ("system", SYSTEM_PROMPT), ("user", "{input}")])

    chain = chat_prompt | llm | output_parser
    response = chain.invoke({"input": text})  # --> str

    return response

# --------------------------------------------------------------------------- #

class PE_LLM(aiko.PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)
        context.set_protocol("llm:0")

        self.detections = None
        self.add_message_handler(self.detection_handler, TOPIC_DETECTIONS)

    def detection_handler(self, aiko, topic, payload_in):
        self.detections = (time.time(), payload_in.split()[1:])

    def process_frame(self, stream, text) -> Tuple[aiko.StreamEvent, dict]:
        if text != "<silence>":
            detections = ""
            if self.detections:
                time_detected, detections = self.detections
                time_now = time.time()
                if time_now > time_detected + 1.0:
                    detections = ""

            self.logger.info(f"Input: {text}")
            response = llm_chain("ollama", text, detections)

        #   topic_out = f"{get_namespace()}/speech"
        #   payload_out = response
        #   aiko.message.publish(topic_out, payload_out)
        else:
            response = text

        return aiko.StreamEvent.OKAY, {"text": response}

# --------------------------------------------------------------------------- #



#!/usr/bin/env python3
#
# Usage
# ~~~~~
# ./xgo_robot.py
#
# To Do
# ~~~~~
# - Wrap as a PipelineElement for audio / video / LLM Pipeline
#   - Can operate independently of "robot_control.py"
#
# - Replace "REAL_ROBOTS" with a mapping of host names to implementations
#   - Command line option: robot_implementation_class
#
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# - Add "period" to "move(direction, stride)" and "turn(speed)"
#
# - Implement (lamda NAME (COMMAND ...) (COMMAND ...)) --> saved as "NAME"
#   - (remove NAME), (run NAME) --> run in background thread
#   - (sleep TIME), (do NAME FROM TO INCREMENT)
#   - Nod head for "yes" or "no"
#
# - Add cv2.set(10, 1) for brightness
# - Acquire CPU% and display on screen
# - Button A (bottom-right) --> "(ml yolo)"
# - Send video images over MQTT as an HL DataSource
# - Send Yolo ML Model inferences over MQTT as an HL DataSource
#
# - Provide logging, especially for syntax errors and run-time Exceptions
# - Determine if "xgolib" can be made asychronous (non-blocking) ?
#
# - Fix event.add_timer_handler(..., immediate=True)
#
# - New menu and button actions
# - LCD screen (message /in)
# - Play sound (message /in)
# - Speech synthesis (message /in)
#
# - Provide camera video stream for off-board viewing and ML processing
#   - QR Code reader --> actions
#   - Aruco tag reader --> actions or location / pose
# - Select ML Model for local video processing (message /in)
#
# - LIDAR: LD06
# - Oak-D Lite camera
# - ROS2 integration

from abc import abstractmethod
import cv2              # pip install opencv-python
from io import BytesIO
import numpy as np
from PIL import Image
import platform
from threading import Thread
import time
import zlib

IS_ROBOT = None
REAL_ROBOTS = ["laika", "oscar"]  # real robot hostnames

def is_robot():
    global IS_ROBOT
    if IS_ROBOT is None:
        hostname = platform.uname()[1]
        IS_ROBOT = hostname in REAL_ROBOTS
    return IS_ROBOT

if is_robot():
    import RPi  # pip install RPi.GPIO
    import spidev
    from key import Button
    import LCD_2inch
    from xgolib import XGO

# --------------------------------------------------------------------------- #

import aiko_services as aiko

_VERSION = 0

ACTOR_TYPE = "xgo_robot"
PROTOCOL = f"{ServiceProtocol.AIKO}/{ACTOR_TYPE}:{_VERSION}"

_LOGGER = aiko.logger(__name__)

if is_robot():
    MODULE_PATH = "aiko_services.examples.xgo_robot.xgo_robot"
else:
    MODULE_PATH = "__main__"

# --------------------------------------------------------------------------- #

ACTIONS = {
  "fall":         1, "stand":           2, "crawl":      3, "circle":       4,
  "step":         5, "squat":           6, "roll":       7, "pitch":        8,
  "yaw":          9, "roll_pitch_yaw": 10, "pee":       11, "sit":         12,
  "beckon":      13, "stretch":        14, "wave":      15, "wiggle_body": 16,
  "wiggle_tail": 17, "sniff":          18, "shake_paw": 19, "arm":         20
}

BATTERY_MONITOR_PERIOD = 10.0  # seconds
DETAIL_XY = (10, 20)
SLEEP_PERIOD = 0.2             # seconds
STATUS_XY = (10, 235)
TOPIC_VIDEO = f"{get_namespace()}/video"

# --------------------------------------------------------------------------- #

class XGORobot(Actor):
    Interface.default("XGORobot", f"{MODULE_PATH}.XGORobotImpl")

    @abstractmethod
    def action(self, value):
        pass

    @abstractmethod
    def arm(self, x, z):  # x: -80 to 155, z: -95 to 155
        pass

    @abstractmethod
    def arm_mode(self, stabilize):  # stabilize: true or false
        pass

    @abstractmethod                                    # pitch: -15 to 15
    def attitude(pitch="nil", roll="nil", yaw="nil"):  # roll:  -20 to 10
        pass                                           # yaw:   -11 to 11

    @abstractmethod
    def body_mode(self, stabilize):  # stabilize: true or false
        pass

    @abstractmethod
    def claw(self, grip):  # grip: 0 (open) to 255 (closed) --> 0% to 100%
        pass

    @abstractmethod                           # direction: "x" or "y"
    def move(self, direction, stride="nil"):  # stride x:  -25 mm to 25 mm
        pass                                  # stride y:  -18 mm to 18 mm

    @abstractmethod
    def reset(self):
        pass

    @abstractmethod
    def screen_detail(self, enabled=None):  # enabled: bool, default: toggle
        pass

    @abstractmethod
    def stop(self):  # stop moving or rotating
        pass

    @abstractmethod
    def terminate(self, immediate=False):
        pass

    @abstractmethod                              # x: -35 (back) to 35 (forward)
    def translation(x="nil", y="nil", z="nil"):  # y: -18 (left) to 18 (right)
        pass                                     # z:  75 (down) to 115 (up)

    @abstractmethod
    def turn(self, speed):  # speed: -100 (clockwise) to 100 degrees / second
        pass

# --------------------------------------------------------------------------- #

class RobotCore:
    def __init__(self, context):
        print(f"MQTT topic: {self.topic_path}")

        self.share["battery"] = -1
        self.share["screen_detail"] = False
        self.share["sleep_period"] = SLEEP_PERIOD
        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"
        self.share["topic_video"] = TOPIC_VIDEO
        self.share["version_firmware"] = "v0"

        self._direction = "x"
        self._pitch = 0
        self._roll = 0
        self._stride = 0
        self._terminated = False
        self._x = 0
        self._y = 0
        self._yaw = 0
        self._z = 0

        self._xgo = None
      # self._button = Button()                   # TODO: Mock later
        self._camera = self._camera_initialize()  # TODO: Optional
        self._screen = None
        self._thread = Thread(target=self._run).start()

    def _run(self):
        fps = 0
        while not self._terminated:
            time_loop = time.time()
            status, image = self._camera.read()
            image = cv2.flip(image, 1)
            image = self._image_to_rgb(image)
            time_process = (time.time() - time_loop) * 1000

            if self._screen:
                self._screen_overlay(image, fps, time_process)  # TODO: Always
                self._screen_show(image)
            self._publish_image(image)

    # GUI robot_control.py button(s) --[MQTT function call]--> xgo_robot.py
        if is_robot():                  # TODO: Mock later
            if self._button.press_c():  # Top-left button
                self.screen_detail()

            if self._button.press_b():  # Bottom-left button
                self.terminate()

        self._sleep()
        fps = int(1 / (time.time() - time_loop))

# --------------------------------------------------------------------------- #

class XGORobotImpl(XGORobot, RobotCore):
    def __init__(self, context):
        context.get_implementation("Actor").__init__(self, context)
        RobotCore.__init__(self, context)

        if is_robot():
            self._xgo = XGO(port="/dev/ttyAMA0", version="xgomini")
            self.share["version_firmware"] = self._xgo.read_firmware()
            self.share["version_xgolib"] = self._xgo.read_lib_version()

            self._button = Button()
            self._screen = self._screen_initialize()
            self._xgo.claw(0)  # Show signs of life !

            event.add_timer_handler(  # TODO: Use laptop battery status
                self._monitor_battery, BATTERY_MONITOR_PERIOD, immediate=True)

    # Review "xgolib": Blocks robot (all threads) until action is completed :(
    def action(self, action_type):
        if action_type in ACTIONS:
# MOCK      self._xgo.action(ACTIONS[action_type])
            payload_out = f"(action {action_type})"
            aiko.message.publish(self.topic_out, payload_out)

    def arm(self, x, z):
        try:
# MOCK      self._xgo.arm(int(x), int(z))
            payload_out = f"(arm {x} {z})"
            aiko.message.publish(self.topic_out, payload_out)
        except:
            pass

    def arm_mode(self, stabilize):
        stabilize = 1 if stabilize == "true" else 0
# MOCK  self._xgo.arm_mode(stabilize)
        payload_out = f"(arm_mode {stabilize})"
        aiko.message.publish(self.topic_out, payload_out)

    def attitude(self, pitch="nil", roll="nil", yaw="nil"):
        try:
            self._pitch = int(pitch)
        except:
            pass

        try:
            self._roll = int(roll)
        except:
            pass

        try:
            self._yaw = int(yaw)
        except:
            pass

# MOCK  self._xgo.attitude(["p","r","y"], [self._pitch, self._roll, self._yaw])
        payload_out = f"(attitude {self._pitch} {self._roll} {self._yaw})"
        aiko.message.publish(self.topic_out, payload_out)

    def body_mode(self, stabilize):
        stabilize = 1 if stabilize == "true" else 0
# MOCK  self._xgo.imu(stabilize)
        payload_out = f"(body_mode {stabilize})"
        aiko.message.publish(self.topic_out, payload_out)

    def _camera_initialize(self):
        camera = cv2.VideoCapture(0)
        camera.set(3, 320)
        camera.set(4, 240)
        return camera

    def claw(self, grip):  # grip: 0% - 100%
        try:
# MOCK      self._xgo.claw(int(grip))
            payload_out = f"(claw {grip})"
            aiko.message.publish(self.topic_out, payload_out)
        except:
            pass

    def _image_to_rgb(self, image):
        b, g, r = cv2.split(image)
        image = cv2.merge((r, g, b))
        return image

    def _monitor_battery(self):
        self.share["battery"] = self._xgo.read_battery()
        self.ec_producer.update("battery", self.share["battery"])
        payload_out = f"(battery {self.share['battery']})"
        aiko.message.publish(self.topic_out, payload_out)

    def move(self, direction, stride="nil"):
        self._direction = "y" if direction == "y" else "x"
        try:
            self._stride = int(stride)
        except:
            pass

# MOCK  self._xgo.move(self._direction, self._stride)
        payload_out = f"(move {self._direction} {self._stride})"
        aiko.message.publish(self.topic_out, payload_out)

    def _publish_image(self, image):
        payload_out = BytesIO()
        np.save(payload_out, image, allow_pickle=True)
        payload_out = zlib.compress(payload_out.getvalue())
        aiko.message.publish(self.share["topic_video"], payload_out)

    def reset(self):
# MOCK  self._xgo.reset()
        self._pitch = self._roll = self._yaw = 0
        self._x = self._y = self._z = 0
        self._direction = "x"
        self._stride = 0
        aiko.message.publish(self.topic_out, "(reset)")

    def screen_detail(self, enabled=None):
        if enabled == None:  # Toggle "screen_detail" enabled
            try:
                enabled = not bool(self.share["screen_detail"])
            except:
                enabled = False
        self.ec_producer.update("screen_detail", enabled)
        payload_out = f"(screen_detail {enabled})"
        aiko.message.publish(self.topic_out, payload_out)

    def _screen_initialize(self):
        screen = None
    #   screen = LCD_2inch.LCD_2inch()
    #   screen.clear()
    #   image = Image.new("RGB", (screen.height, screen.width), "black")
    #   screen.ShowImage(image)
        return screen

    def _screen_overlay(self, image, fps, time_process):
        status = f"{self.share['battery']}%  {time_process:.01f} ms  {fps} FPS"
        cv2.putText(image, status, STATUS_XY, 0, 0.7, (255, 255, 255), 2)

        if self.share["screen_detail"]:
            detail = f"{self.topic_path}"
            cv2.putText(image, detail, DETAIL_XY, 0, 0.7, (255, 255, 255), 2)

    def _screen_show(self, image):
        image = Image.fromarray(image)
# MOCK  self._screen.ShowImage(image)

    def _sleep(self, period=None):
        if not period:
            try:
                period = float(self.share["sleep_period"])
            except:
                period = SLEEP_PERIOD
        time.sleep(period)

    def stop(self):
# MOCK  self._xgo.stop()
        self._direction = "x"
        self._stride = 0
        payload_out = f"(stop)"
        aiko.message.publish(self.topic_out, payload_out)

    def terminate(self, immediate=False):
# MOCK  self._xgo.reset()
# MOCK  self._xgo.claw(128)
        aiko.process.terminate()
        self._terminated = True

    def translation(self, x="nil", y="nil", z="nil"):
        try:
            self._x = int(x)
        except:
            pass

        try:
            self._y = int(y)
        except:
            pass

        try:
            self._z = int(z)
        except:
            pass

# MOCK  self._xgo.translation(["x", "y", "z"], [self._x, self._y, self._z])
        payload_out = f"(attitude {self._x} {self._y} {self._z})"
        aiko.message.publish(self.topic_out, payload_out)

    def turn(self, speed):  # speed: -100 (clockwise) to 100 degrees / second
        try:
# MOCK      self._xgo.turn(int(speed))
            payload_out = f"(turn {speed})"
            aiko.message.publish(self.topic_out, payload_out)
        except:
            pass

# --------------------------------------------------------------------------- #

if __name__ == "__main__":
    init_args = actor_args(ACTOR_TYPE, protocol=PROTOCOL)
    xgo_robot = compose_instance(XGORobotImpl, init_args)
    aiko.process.run()

# --------------------------------------------------------------------------- #



#!/usr/bin/env python3
#
# Manage Pipelines consisting of PipelineElements (Actors or Services)
#
# Usage
# ~~~~~
# DEFINITION=pipeline_definition.json
# aiko_pipeline create  [--name $PIPELINE_NAME] $DEFINITION
# aiko_pipeline destroy $PIPELINE_NAME
#
# aiko_pipeline create $DEFINITION --log_level debug  \
#   --stream_id 1 --frame_data "(argument_name: argument_value ...)"
#
# NAMESPACE=AIKO
# HOST=localhost
# PID=`ps ax | grep aiko_pipeline | grep -v grep | tr -s " " | cut -d" " -f1-2`
# SID=1
# TOPIC=$NAMESPACE/$HOST/$PID/$SID/in
#
# mosquitto_pub -h $HOST -t $TOPIC -m "(create_stream 1)"
# mosquitto_pub -h $HOST -t $TOPIC -m "(process_frame (stream_id: 1) (a: 0))"
# mosquitto_pub -h $HOST -t $TOPIC -m "(destroy_stream 1)"
#
# Definition
# ~~~~~~~~~~
# "graph": [
#   "(PE_0 PE_1)",
#   "(PE_0 PE_1 (PE_2 PE_1))",
#   "(PE_0 (PE_1 (PE_3 PE_5)) (PE_2 (PE_4 PE_5)))"
# ]
#
# PipelineElement function argument name mapping example ...
# - PE_1 outputs "a" and PE_2 outputs "b"
# - Whilst PE_3 expects inputs "x" and "y"
#
# "graph": [
#   "(PE_0 (PE_1 PE_3 (a: x)) (PE_2 PE_3 (b: y)))"
# ]
#
# Important
# ---------
# - Pipeline is-a PipelineElement and a Category (of PipelineElements)
#
# - Pipeline Definition Avro schema is hard-coded into this source file ...
#   - To ensure that the Avro schema matches the source code implementation
#   - Avoid having to refer to a critical file somewhere in the filesystem
#
# - PipelineDefinition inputs
#   - Program language (Python) data structure is the gold standard
#   - Avro parsed and validated JSON or S-Expressions
#     - Consider using FastAvro (CPython)
#   - GraphQL
#
# Resources
# ~~~~~~~~~
# - Avro 1.9.1 specification
#   - https://avro.apache.org/docs/1.9.1/spec.html
#
#   - https://www.perfectlyrandom.org/2019/11/29/handling-avro-files-in-python
#     - https://github.com/linkedin/python-avro-json-serializer
#     - https://marcosschroh.github.io/python-schema-registry-client
#
# To Do
# ~~~~~
# * CLI: pipeline.py show <service_filter>
# * CLI: pipeline.py get <service_filter> <parameter_name>  # or wildcard "*"
# * CLI: pipeline.py set <service_filter> <parameter_name> <parameter_value>
#
# * Should "stream_start()" and "stream_stop()" should return success/failure ?
#   - Yes and the "swag" should be just like "process_frame()"
#
# * Handle list of sub-graphs for multiple sources of different data types
#   - With StateMachine support for dynamic Graph routing / traversal
#       "graph: [
#         "(PE_0 default:)",
#         "(PE_0 streaming: (PE_1 PE_3) (PE_2 PE_3))"
#       ]
#
# - Collect "local" and "remote" into "deployment" configuration structure
#
# - Validate function inputs and outputs against Pipeline Definition
#
# - Pipeline CLI option to be the LifeCycleManager and recursively create both
#   local and *remote* Pipeline / PipelineElements
#
# - pipeline_2022.py ...
#   - ServiceDefinition: map-out, map-in and name_mapping
#   - PipelineElementDefinition(ServiceDefinition): service_level_agreement
#   - PipelineDefinition(PipelineElementDefinition): edges: List[Tuple[PE, PE]]
#
# - For local PipelineElements, better module loading, etc
# - Handle remote PipelineElements
#   - Improve Service/ActorDiscovery to implement dynamic Service/ActorProxy
#     - Until Service/Actor is discovered, then dynamic Proxy is a default
#     - When Service/Actor is discovered, then dynamic Proxy is updated
#     - When Service/Actor vanishes, then dynamic Proxy returns to default
#     - Provide "absent" / "ready" status
#   - Collect "topic_path", etc into a "service_filter" structure

from abc import abstractmethod
import avro.schema
from avro_validator.schema import Schema
import click
import copy
from collections import OrderedDict  # All OrderedDict operations are O(1)
from dataclasses import dataclass, asdict
from enum import Enum
import json
import os
import queue
import threading
from threading import Thread, current_thread
import time
import traceback
from typing import Any, Dict, List, Tuple

from aiko_services.main import *
from aiko_services.main.transport import *
from aiko_services.main.utilities import *

__all__ = [
    "Pipeline", "PipelineElement", "PipelineElementImpl", "PipelineImpl",
    "PROTOCOL_PIPELINE"
]

_VERSION = 0

ACTOR_TYPE_PIPELINE = "pipeline"
ACTOR_TYPE_ELEMENT = "pipeline_element"
PROTOCOL_PIPELINE =  f"{ServiceProtocol.AIKO}/{ACTOR_TYPE_PIPELINE}:{_VERSION}"
PROTOCOL_ELEMENT =  f"{ServiceProtocol.AIKO}/{ACTOR_TYPE_ELEMENT}:{_VERSION}"

_GRACE_TIME = 60  # seconds
_LOGGER = aiko.logger(__name__)

# --------------------------------------------------------------------------- #

class DeployType(Enum):
    LOCAL = "local"
    REMOTE = "remote"

@dataclass
class PipelineDefinition:
#   comment: str  # Specified in Avro schema and discarded by parser
    version: int
    name: str
    runtime: str
    graph: List[str]
    parameters: Dict  # Optional field, default: {}
    elements: List

@dataclass
class PipelineElementDefinition:
#   comment: str  # Specified in Avro schema and discarded by parser
    name: str     # unique name
    input: Dict[str, str]
    output: Dict[str, str]
    parameters: Dict  # Optional field, default: {}
    deploy: Dict

@dataclass
class PipelineElementDeployLocal:
    class_name: str  # optional field, default: name
    module: str

@dataclass
class RemoteServiceFilter:
    topic_path: str
    name: str
    owner: str
    protocol: str
    transport: str
    tags: str

@dataclass
class PipelineElementDeployRemote:
    module: str  # TODO: Is this really needed ?  Probably not !
    service_filter: RemoteServiceFilter

# --------------------------------------------------------------------------- #

class PipelineGraph(Graph):
    def __init__(self, head_nodes=None):
        super().__init__(head_nodes)

    def add_element(self, element):
        self.add(element)
        element.predecessors = OrderedDict()

    @property
    def element_count(self):
        return len(self._graph)

    @classmethod
    def get_element(cls, node):
        element = node.element
        if element.__class__.__name__ != "ServiceRemoteProxy":
            name = element.__class__.__name__
            local = element.is_local()
            lifecycle = element.share["lifecycle"]
        else:
            name = node.name
            local = False
            lifecycle = "ready"
        return element, name, local, lifecycle

# TODO: Work-in-progress
# - Pass 1: Validate_inputs_by_name_and_type()
#   - Each input NAME and TYPE provided by exactly one parent predecessor
#   - Each input NAME and TYPE provided by exactly one predecessor (recursive)
#
# - Pass 2: Validate_inputs_by_position_and_type()
#   - Each input TYPE provided exactly in order by parent predecessor
#   - Each input TYPE provided by exactly one predecessor (recursive) ?
#
# - Pass 3: Perform mapping first, then ...
#   - Check mapping output --> input: NAME and TYPE must match
#     "graph": ["(A (B D (d0: d1)) (C D))"],
#
# - Provide mapping clues
# - Flexible or Strict (exact match predecessor output --> successor input)
# - Reflect on PipelineElement function signatures to create definition.elements

    def validate_inputs(self, inputs, predecessors, checked=None, strict=False):
        checked = checked if checked else []  # each predecessor checked once
        for predecessor in predecessors.values():
            if predecessor not in checked:
                checked.append(predecessor)
                predecessor_outputs = predecessor.element.definition.output
                for input in inputs:
                    for predecessor_output in predecessor_outputs:
                        if predecessor_output["name"] == input["name"]:
                            input["found"] += 1
                if not strict:
                    inputs, checked = self.validate_inputs(
                                      inputs, predecessor.predecessors, checked)
        return inputs, checked

    def validate_mapping(self, map_in_nodes, element_name, input):
        valid_mappings = []
        if element_name in map_in_nodes:
            map_in_node = map_in_nodes[element_name]
            for predecessor_name, mapping in map_in_node.items():
                if input["name"] in mapping.values():
                    valid_mappings.append((predecessor_name, mapping))
        return valid_mappings

    def validate(self, pipeline_definition, strict=False):
        for node in self:
            element, element_name, _, _ = PipelineGraph.get_element(node)
            element_inputs = element.definition.input
            element_inputs = [{**item, "found": 0} for item in element_inputs]

            if element_name not in self._head_nodes:
                predecessors = node.predecessors
                if len(predecessors) == 0:  # Don't worry about graph head nodes
                    continue
                inputs, _ = self.validate_inputs(
                                element_inputs, predecessors, strict)
                for input in inputs:
                    try_map_in_out = False
                    diagnostic = f"PipelineElement {element_name}: input \"{input['name']}\" not produced by any "
                    if strict and input["found"] != 1:
                        diagnostic += "immediate predecessor PipelineElement"
                        try_map_in_out = True
                    #   print(f"{diagnostic}")  # TODO: Ensure this is right
                    elif input["found"] == 0:
                        diagnostic += "previous PipelineElements"
                        try_map_in_out = True

                    if try_map_in_out:
                        map_in_nodes = pipeline_definition.map_in_nodes
                        if not self.validate_mapping(
                            map_in_nodes, element_name, input):
                            pass
                        #   print(f"{diagnostic}")  # TODO: Ensure this is right

            for successor_name in node.successors:
                successor = self.get_node(successor_name)
                successor.predecessors[element_name] = node

# --------------------------------------------------------------------------- #
# PipelineElement: service_level_agreement: low_latency, etc

class PipelineElement(Actor):
    Interface.default("PipelineElement",
        "aiko_services.main.pipeline.PipelineElementImpl")

    @abstractmethod
    def create_frame(self, stream, frame_data):
        pass

    @abstractmethod
    def create_frames(
        self, stream, frame_generator, frame_id=FIRST_FRAME_ID, rate=None):
        pass

    @abstractmethod
    def get_parameter(self, name, default=None, use_pipeline=True):
        pass

    @abstractmethod
    def get_stream(self):
        pass

    @classmethod
    def is_local(cls):
        return True

    @abstractmethod
    def my_id(self, all=False):
        pass

# PipelineElement developer must provide a process_frame() implementation

    @abstractmethod
    def process_frame(self, stream, **kwargs) -> Tuple[StreamEvent, dict]:
        """
        Returns "status" indicating success or failure of processing the frame
        """
        pass

    @abstractmethod
    def start_stream(self, stream, stream_id):
        pass

    @abstractmethod
    def stop_stream(self, stream, stream_id):
        pass

class PipelineElementImpl(PipelineElement):
    def __init__(self, context):
        self.definition = context.get_definition()
        self.pipeline = context.get_pipeline()
        self.is_pipeline = (self.pipeline == None)
        if context.protocol == "*":
            context.set_protocol(
                PROTOCOL_PIPELINE if self.is_pipeline else PROTOCOL_ELEMENT)
        context.get_implementation("Actor").__init__(self, context)

        self.share["source_file"] = f"v{_VERSION}⇒ {__file__}"
        self.share.update(self.definition.parameters)
    # TODO: Fix Aiko Dashboard / EC_Producer incorrectly updates this approach
    #   self.share["parameters"] = self.definition.parameters  # TODO

    def create_frame(self, stream, frame_data, frame_id=None):
        frame_id = frame_id if frame_id else stream.frame_id
        stream_copy = Stream(stream.stream_id, frame_id,
            stream.parameters, stream.queue_response,
            stream.state, stream.topic_response)
        self.pipeline.create_frame(stream_copy, frame_data)

# Optional "frame_generator" starts during "create_stream() --> start_stream()"

    def create_frames(
        self, stream, frame_generator, frame_id=FIRST_FRAME_ID, rate=None):

        thread_args = (stream, frame_generator, int(frame_id), rate)
        Thread(target=self._create_frames_generator,
            args=thread_args, daemon=True).start()

# TODO: For "rate" measure time since last frame to be more accurate
# FIX:  For "rate" check "rate=0" (fills mailbox) versus "rate=None" ?

    def _create_frames_generator(self, stream, frame_generator, frame_id, rate):
        try:
            self.pipeline._enable_thread_local(
                "_create_frames_generator", stream.stream_id, frame_id)
            stream, frame_id = self.get_stream()

            while stream.state == StreamState.RUN:
                try:
                    stream_event, frame_data = frame_generator(stream, frame_id)
                except Exception as exception:
                    self.logger.error(
                        "Exception in pipeline_element._create_frames() --> "  \
                        "frame_generator()")
                    stream_event = StreamEvent.ERROR
                    frame_data = {"diagnostic": traceback.format_exc()}

                stream.state = self.pipeline._process_stream_event(
                    self.name, stream_event, frame_data)

                if stream.state == StreamState.RUN:
                # TODO: Check "isinstance(frame_data, dict)"
                    if frame_data:
                        self.create_frame(stream, frame_data, frame_id)
                    if rate:
                        time.sleep(1.0 / rate)
                    frame_id += 1
                    self.pipeline.thread_local.frame_id = frame_id
        finally:
            self.pipeline._disable_thread_local("_create_frames_generator")

    def get_parameter(self, name, default=None, use_pipeline=True):
    # TODO: During process_frame(), stream parameters should be updated
    #       in self.share, just like PipelineDefinition parameters.
    #       Note: Consider the performance implications when doing this !

        value = None
        found = False

        element_parameter_name = f"{self.definition.name}.{name}"
        stream_parameters = self._get_stream_parameters()

        if element_parameter_name in stream_parameters:
            value = stream_parameters[element_parameter_name]
            found = True
        elif name in self.definition.parameters:
            if name in self.share:
                value = self.share[name]
                found = True

    # TODO: Should also allow Pipeline parameters to be updated

        if not found and use_pipeline and not self.is_pipeline:
            if name in stream_parameters:
                value = stream_parameters[name]
                found = True
            elif name in self.pipeline.definition.parameters:
                if name in self.pipeline.share:
                    value = self.pipeline.share[name]
                    found = True
        if not found and default is not None:
            value = default  # Note: "found" is deliberately left as False
        return value, found

    def get_stream(self):
        return self.pipeline.get_stream()

    def _get_stream_parameters(self):
        parameters = {}
        stream, _ = self.get_stream()
        if stream:
            parameters = stream.parameters
        return parameters

    def my_id(self, all=False):
        name = self.name if all else ""
        stream, frame_id = self.get_stream()
        return f"{name}<{stream.stream_id}:{frame_id}>"

    def start_stream(self, stream, stream_id):
        return StreamEvent.OKAY, None

    def stop_stream(self, stream, stream_id):
        return StreamEvent.OKAY, None

# --------------------------------------------------------------------------- #

class Pipeline(PipelineElement):
    Interface.default("Pipeline", "aiko_services.main.pipeline.PipelineImpl")

    @abstractmethod
    def create_stream(self, stream_id, parameters=None, grace_time=_GRACE_TIME,
        queue_response=None, topic_response=None):
        pass

    @abstractmethod
    def destroy_stream(self, stream_id):
        pass

    @abstractmethod
    def parse_pipeline_definition(cls, pipeline_definition_pathname):
        pass

    @abstractmethod
    def process_frame_response(
        self, stream, frame_data) -> Tuple[StreamEvent, dict]:
        pass

    @abstractmethod
    def set_parameter(self, stream_id, name, value):
        pass

class PipelineImpl(Pipeline):
    DEPLOY_TYPE_LOOKUP = {
        DeployType.LOCAL.value: PipelineElementDeployLocal,
        DeployType.REMOTE.value: PipelineElementDeployRemote
    }
    DEPLOY_TYPE_LOCAL_NAME = PipelineElementDeployLocal.__name__
    DEPLOY_TYPE_REMOTE_NAME = PipelineElementDeployRemote.__name__

    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)
        print(f"MQTT topic: {self.topic_in}")

        self.share["definition_pathname"] = context.definition_pathname
        self.share["lifecycle"] = "waiting"
        self.remote_pipelines = {}  # Service name --> PipelineElement name
        self.services_cache = None

        self.stream_leases = {}
        self.thread_local = threading.local()  # See _enable_thread_local()

        self.pipeline_graph = self._create_pipeline_graph(context.definition)
        self.share["element_count"] = self.pipeline_graph.element_count
        self.share["streams"] = 0
        self.share["streams_frames"] = 0
        self._update_lifecycle_state()

    # TODO: Better visualization of the Pipeline / PipelineElements details
        if False:
            print(f"Pipeline: {self.pipeline_graph.nodes()}")
            for node in self.pipeline_graph:
                print(f"PipelineElement: {node.name}")

        event.add_timer_handler(self._status_update_timer, 1.0)

    def _update_lifecycle_state(self):
        pe_lifecycles = []
        for node in self.pipeline_graph:
            _, _, _, lifecycle = PipelineGraph.get_element(node)
            pe_lifecycles.append(lifecycle == "ready")
        lifecycle = "ready" if all(pe_lifecycles) else "waiting"
        self.ec_producer.update("lifecycle", lifecycle)

    def _status_update_timer(self):
        streams_frames = 0
        for stream_lease in self.stream_leases.values():
            streams_frames += len(stream_lease.stream.frames)

        self.ec_producer.update("streams", len(self.stream_leases))
        self.ec_producer.update("streams_frames", streams_frames)

    def _add_node_properties(self, node_name, properties, predecessor_name):
        definition = self.definition

        if node_name not in definition.map_in_nodes:
            definition.map_in_nodes[node_name] = {}
        definition.map_in_nodes[node_name][predecessor_name] = properties

        if predecessor_name not in definition.map_out_nodes:
            definition.map_out_nodes[predecessor_name] = {}
        definition.map_out_nodes[predecessor_name][node_name] = properties

# Pipeline current "stream" and "frame_id" are thread-local variables
# Valid for create_stream(), process_frame() and destroy_stream() on main thread
# Valid for PipelineElement._create_frames_generator() thread
#
# Always use ...
#     try:
#         self._enable_thread_local("function_name", stream_id, frame_id)
#         stream, frame_id = self.get_stream()
#     finally:
#         self._disable_thread_local("function_name")

    def _enable_thread_local(self, function_name, stream_id, frame_id=None):
        stream = getattr(self.thread_local, "stream", None)
        assert not stream, "self.thread_local.stream must not be assigned"
        self.thread_local.stream = self.stream_leases[stream_id].stream
        if frame_id:
            self.thread_local.frame_id = frame_id
        else:
            self.thread_local.frame_id = self.thread_local.stream.frame_id
    #   self.logger.warning(f"Enable:  {function_name}: {self.my_id()}")  # useful

    def _disable_thread_local(self, function_name):
        stream = self.thread_local.stream
        assert stream, "self.thread_local.stream must be assigned"
    #   self.logger.warning(f"Disable: {function_name}: {self.my_id()}")  # useful
        self.thread_local.stream = None
        self.thread_local.frame_id = None

    def create_frame(self, stream_dict, frame_data):
        if isinstance(stream_dict, Stream):
            stream_dict = stream_dict.as_dict()
        self._post_message(
            ActorTopic.IN, "process_frame", [stream_dict, frame_data])

    @classmethod
    def create_pipeline(cls, definition_pathname, pipeline_definition,
        name, stream_id, stream_parameters, frame_id, frame_data, grace_time,
        queue_response=None):

        name = name if name else pipeline_definition.name

        init_args = pipeline_args(name,
            protocol=PROTOCOL_PIPELINE,
            definition=pipeline_definition,
            definition_pathname=definition_pathname
        )
        pipeline = compose_instance(PipelineImpl, init_args)

        stream_dict = {"frame_id": int(frame_id), "parameters": {}}

        if stream_id is not None:
            stream_dict["stream_id"] = stream_id
            stream_parameters = dict(stream_parameters)
            pipeline.create_stream(stream_id, parameters=stream_parameters,
                grace_time=grace_time, queue_response=queue_response)

        if frame_data is not None:
            function_name, arguments = parse(f"(process_frame {frame_data})")
            if len(arguments):
                pipeline.create_frame(stream_dict, arguments[0])
            else:
                raise SystemExit(
                    f"Error: Frame data must be provided")

        return pipeline

    def _create_pipeline_graph(self, definition):
        header = f"Error: Creating Pipeline: {definition.name}"

        if len(definition.elements) == 0:
            self._error_pipeline(header,
                "PipelineDefinition: Doesn't define any PipelineElements")

        definition.map_in_nodes = {}
        definition.map_out_nodes = {}
        node_heads, node_successors = Graph.traverse(
            definition.graph, self._add_node_properties)
        pipeline_graph = PipelineGraph(node_heads)
    #   self.parameters = definition.parameters  # TODO: Use it or lose it !

        for pipeline_element_definition in definition.elements:
            element_instance = None
            element_name = pipeline_element_definition.name
            if element_name not in node_successors:
                print(f'Warning: Skipping PipelineElement {element_name}: '
                      f'Not used within the "graph" definition')
                continue
            deploy_definition = pipeline_element_definition.deploy
            deploy_type_name = type(deploy_definition).__name__

            if deploy_type_name == PipelineImpl.DEPLOY_TYPE_LOCAL_NAME:
                class_name = pipeline_element_definition.deploy.class_name
                element_class = self._load_element_class(
                    deploy_definition.module, class_name, header)

            # TODO: Is element_name is correct for remote case

            if deploy_type_name == PipelineImpl.DEPLOY_TYPE_REMOTE_NAME:
                element_class = PipelineRemoteAbsent
                service_name = deploy_definition.service_filter["name"]
                if service_name not in self.remote_pipelines:
                    self.remote_pipelines[service_name] = element_name
                else:
                    self._error_pipeline(header,
                        f"PipelineDefinition: PipelineElement {element_name}: "
                        f"re-uses remote service_filter name: {service_name}")
                if not self.services_cache:
                    self.services_cache = services_cache_create_singleton(self)
                service_filter = ServiceFilter.with_topic_path(
                    **deploy_definition.service_filter)
                self.services_cache.add_handler(
                    self._pipeline_element_change_handler, service_filter)

            if not element_class:
                self._error_pipeline(header,
                    f"PipelineDefinition: PipelineElement type unknown: "
                    f"{deploy_type_name}")

            init_args = pipeline_element_args(element_name,
                definition=pipeline_element_definition,
                pipeline=self
            )
            element_instance = compose_instance(element_class, init_args)
            element_instance.parameters = pipeline_element_definition.parameters

            element = Node(
                element_name, element_instance, node_successors[element_name])
            pipeline_graph.add_element(element)

        pipeline_graph.validate(definition)
        return pipeline_graph

# TODO: Consider refactoring Stream into "stream.py" ?

    def create_stream(self, stream_id, parameters=None, grace_time=_GRACE_TIME,
        queue_response=None, topic_response=None):

        if queue_response and topic_response:
            self.logger.error(
                f"Create stream: use either queue_response or topic_response")
            return False

    # TODO: Implement limit on delayed post message
        if self.share["lifecycle"] != "ready":
            arguments = [
                stream_id, parameters, grace_time,
                queue_response, topic_response]
            self._post_message(
                ActorTopic.IN, "create_stream", arguments, delay=1.0)
            return False

        if stream_id in self.stream_leases:
            self.logger.error(f"Create stream: {stream_id} already exists")
            return False

        self.logger.debug(f"Create stream: {self.name}<{stream_id}>")
        stream_lease = Lease(int(grace_time), stream_id,
            lease_expired_handler=self.destroy_stream)
        stream_lease.stream = Stream(
            stream_id=stream_id,
            parameters=parameters if parameters else {},
            queue_response=queue_response,
            topic_response=topic_response)
        self.stream_leases[stream_id] = stream_lease

        try:
            self._enable_thread_local("create_stream", stream_id)
            stream, _ = self.get_stream()

            for node in self.pipeline_graph:
                element, element_name, local, _ =  \
                    PipelineGraph.get_element(node)
                if local:  ## Local element ##
                    try:
                        stream_event, diagnostic = element.start_stream(
                            stream, stream_id)
                    except Exception as exception:
                        self.logger.error("Exception in "  \
                          "pipeline.create_stream() --> start_stream()")
                        stream_event = StreamEvent.ERROR
                        diagnostic = {"diagnostic": traceback.format_exc()}

                    stream_state = self._process_stream_event(
                        element_name, stream_event, diagnostic)
                else:  ## Remote element ##
                # TODO: Consider using "topic_response=self.topic_control"
                    element.create_stream(
                        stream_id, parameters, grace_time, None, self.topic_in)
        finally:
            self._disable_thread_local("create_stream")
        return True

    def destroy_stream(self, stream_id, use_thread_local=True):
        if stream_id not in self.stream_leases:
            return False

        try:
            if use_thread_local:
                self._enable_thread_local("destroy_stream", stream_id)
            stream, _ = self.get_stream()
            self.logger.debug(f"Destroy stream: {self.name}<{stream_id}>")

            for node in self.pipeline_graph:
                element, element_name, local, _ =  \
                    PipelineGraph.get_element(node)
                if local:  ## Local element ##
                    try:
                        stream_event, diagnostic = element.stop_stream(
                            stream, stream_id)
                    except Exception as exception:
                        self.logger.error("Exception in "  \
                          "pipeline.destroy_stream() --> stop_stream()")
                        stream_event = StreamEvent.ERROR
                        diagnostic = {"diagnostic": traceback.format_exc()}

                    stream_state = self._process_stream_event(element_name,
                        stream_event, diagnostic, in_destroy_stream=True)
                else:  ## Remote element ##
                    element.destroy_stream(stream_id)
        finally:
            if use_thread_local:
                self._disable_thread_local("destroy_stream")

        stream_lease = self.stream_leases[stream_id]
        del self.stream_leases[stream_id]
        return True

    def _error_pipeline(self, header, diagnostic):
        PipelineImpl._exit(header, diagnostic)

    @classmethod
    def _exit(cls, header, diagnostic):
        complete_diagnostic = f"{header}\n{diagnostic}"
        _LOGGER.error(complete_diagnostic)
        raise SystemExit(-1)

    def get_stream(self):   # See _enable_thread_local()
        stream = self.thread_local.stream
        assert stream, "self.thread_local.stream must be assigned"
        return stream, self.thread_local.frame_id

    def _load_element_class(self,
        module_descriptor, element_name, header):

        diagnostic = None
        stack_traceback = ""
        try:
            module = load_module(module_descriptor)
            element_class = getattr(module, element_name)
        except FileNotFoundError:
            diagnostic = "found"
        except Exception:
            diagnostic = "loaded"
            stack_traceback = "\n" + str(traceback.format_exc())
        if diagnostic:
            self._error_pipeline(header,
                f"PipelineDefinition: PipelineElement {element_name}: "
                f"Module {module_descriptor} could not be {diagnostic}"
                f"{stack_traceback}")
        return element_class

    @classmethod
    def parse_pipeline_definition(cls, pipeline_definition_pathname):
        CLASS_NAME_FIELD = "class_name"
        COMMENT_FIELD = "#"
        PARAMETERS_FIELD = "parameters"
        header = f"Error: Parsing PipelineDefinition: {pipeline_definition_pathname}"
        try:
            pipeline_definition_dict = json.load(
                open(pipeline_definition_pathname, "r"))
            PIPELINE_DEFINITION_SCHEMA.validate(pipeline_definition_dict)
        except ValueError as value_error:
            PipelineImpl._exit(header, value_error)

        if COMMENT_FIELD in pipeline_definition_dict:  # discard
            del pipeline_definition_dict[COMMENT_FIELD]
        if PARAMETERS_FIELD not in pipeline_definition_dict:  # optional
            pipeline_definition_dict[PARAMETERS_FIELD] = {}

        try:
            pipeline_definition = PipelineDefinition(**pipeline_definition_dict)
        except TypeError as type_error:
            PipelineImpl._exit(header, type_error)

        if pipeline_definition.version != PIPELINE_DEFINITION_VERSION:
            PipelineImpl._exit(header,
                f"PipelineDefinition: Version must be 0, "
                f"but is {pipeline_definition.version}")

        if pipeline_definition.runtime != "python":
            PipelineImpl._exit(header,
                f'PipelineDefinition: Runtime must be "python", '
                f'but is "{pipeline_definition.runtime}"')

        element_definitions = []
        for element_fields in pipeline_definition.elements:
            if COMMENT_FIELD in element_fields:  # discard
                del element_fields[COMMENT_FIELD]
            if PARAMETERS_FIELD not in element_fields:  # optional
                element_fields[PARAMETERS_FIELD] = {}

            try:
                element_definition = PipelineElementDefinition(**element_fields)
            except TypeError as type_error:
                PipelineImpl._exit(header,
                    f"PipelineDefinition: PipelineElement {type_error}")

            if len(element_definition.deploy.keys()) != 1:
                PipelineImpl._exit(header,
                    f"PipelineDefinition: PipelineElement "
                    f"{element_definition.name} must be either local or remote")
            deploy_type = list(element_definition.deploy.keys())[0]

            if deploy_type in PipelineImpl.DEPLOY_TYPE_LOOKUP:
                pipeline_element_deploy_type =  \
                    PipelineImpl.DEPLOY_TYPE_LOOKUP[deploy_type]
            else:
                PipelineImpl._exit(header,
                    f"PipelineDefinition: PipelineElement "
                    f"{element_definition.name}: "
                    f"Unknown Pipeline deploy type: {deploy_type}")

            element_deploy_fields = element_definition.deploy[deploy_type]

            if deploy_type == DeployType.LOCAL.value:
                if CLASS_NAME_FIELD not in element_deploy_fields:  # optional
                    element_deploy_fields[CLASS_NAME_FIELD] =  \
                        element_definition.name

            element_definition.deploy =  \
                pipeline_element_deploy_type(**element_deploy_fields)

            element_definitions.append(element_definition)

        pipeline_definition.elements = element_definitions

        _LOGGER.info(
            f"PipelineDefinition parsed: {pipeline_definition_pathname}")
        return(pipeline_definition)

    def _pipeline_element_change_handler(self, command, service_details):
        if command in ["add", "remove"]:
            print(f"Pipeline change: ({command}: {service_details[0:2]} ...)")
            topic_path = f"{service_details[0]}/in"
            service_name = service_details[1]
            element_name = self.remote_pipelines[service_name]
            node = self.pipeline_graph.get_node(element_name)
            element_definition = node.element.definition

            if command == "add":
                header = f"Error: Updating Pipeline: {element_definition.name}"
            # TODO: Don't create another PipelineElement Service !
            # TODO: Shouldn't need to load "element_definition.deploy.module"
                element_class = self._load_element_class(
                    element_definition.deploy.module, element_name, header)

            if command == "remove":
                element_class = PipelineRemoteAbsent

            init_args = pipeline_element_args(element_name,
                definition=element_definition,
                pipeline=self
            )
            element_instance = compose_instance(element_class, init_args)
            if command == "add":
                element_instance = get_actor_mqtt(
                    topic_path, PipelineRemoteFound)
                element_instance.definition = element_definition
            node._element = element_instance

            self._update_lifecycle_state()
            print(f"Pipeline update: --> {element_name} proxy")

    def process_frame(
        self, stream_dict, frame_data) -> Tuple[StreamEvent, dict]:

        return self._process_frame_common(stream_dict, frame_data, True)

    def process_frame_response(
        self, stream_dict, frame_data) -> Tuple[StreamEvent, dict]:

        return self._process_frame_common(stream_dict, frame_data, False)

    def _process_frame_common(self, stream_dict, frame_data_in, new_frame)  \
        -> Tuple[StreamEvent, dict]:

        graph, stream =  \
            self._process_initialize(stream_dict, frame_data_in, new_frame)
        if graph is None:
            return False

        try:
            self._enable_thread_local("process_frame", stream.stream_id)
            stream, _ = self.get_stream()
            frame = stream.frames[stream.frame_id]
            metrics = self._process_metrics_initialize(frame)

            definition_pathname = self.share["definition_pathname"]
            element_name = None
            frame_complete = True
            frame_data_out = {} if new_frame else frame_data_in

            for node in graph:
                if stream.state == StreamState.ERROR:
                    break
                # TODO: Is element_name is correct for all error diagnostics
                element, element_name, local, _ =  \
                    PipelineGraph.get_element(node)
                header = f'Error: Invoking Pipeline "{definition_pathname}": ' \
                         f'PipelineElement "{element_name}": process_frame()'

                inputs = self._process_map_in(
                    header, element, element_name, frame.swag)

                stream_event = StreamEvent.OKAY
                try:
                    if local:  ## Local element ##
                        start_time = time.time()

                        try:
                            stream_event, frame_data_out =  \
                                element.process_frame(stream, **inputs)
                        except Exception as exception:
                            self.logger.error(
                                "Exception in pipeline.process_frame() --> "  \
                                "frame_generator()")
                            stream_event = StreamEvent.ERROR
                            frame_data_out = {
                                "diagnostic": traceback.format_exc()}

                        stream.state = self._process_stream_event(
                            element_name, stream_event, frame_data_out)
                    #   TODO: Test "stream.state" before continuing
                        self._process_map_out(element_name, frame_data_out)
                        self._process_metrics_capture(  # TODO: Move up ?
                            metrics, element.name, start_time)
                        frame.swag.update(frame_data_out)
                    else:  ## Remote element ##
                        frame_complete = False
                        frame_data_out = {}
                        frame.paused_pe_name = node.name
                        remote_stream = {
                            "stream_id": stream.stream_id,
                            "frame_id": stream.frame_id
                        }
                        element.process_frame(remote_stream, **inputs)
                        break  # process_frame_response() --> continue graph
                except Exception as exception:
                    self._error_pipeline(header, traceback.format_exc())

            if frame_complete:
                stream_info = {
                    "stream_id": stream.stream_id,
                    "frame_id": stream.frame_id}
                if stream.queue_response:
                    stream.queue_response.put((stream_info, frame_data_out))
                elif stream.topic_response:
                    actor = get_actor_mqtt(stream.topic_response, Pipeline)
                    actor.process_frame_response(stream_info, frame_data_out)
                else:
                    payload = generate(
                        "process_frame", (stream_info, frame_data_out))
                    aiko.message.publish(self.topic_out, payload)

        finally:
            if frame_complete:
                del stream.frames[stream.frame_id]
            self._disable_thread_local("process_frame")
        return True

    def _process_initialize(self, stream_dict, frame_data_in, new_frame):
        frame = None
        graph = None
        stream = Stream()
        stream.update(stream_dict)

        if not isinstance(frame_data_in, dict):
            header = f"Process frame <{stream.stream_id}:{stream.frame_id}>"
            self.logger.warning(f"{header}: frame data must be a dictionary")
            return None, None

        stream_id = stream.stream_id
        if stream_id == DEFAULT_STREAM_ID:
            if DEFAULT_STREAM_ID not in self.stream_leases:
                self.create_stream(
                    DEFAULT_STREAM_ID, parameters=stream.parameters)

        frame_id = stream.frame_id
        if stream_id not in self.stream_leases:
            self.logger.warning(
                f"Process frame <{stream_id}:{frame_id}>: stream not found")
        else:
            stream_lease = self.stream_leases[stream_id]
            stream_lease.extend()
            stream_lease.stream.update({"frame_id": frame_id})
            stream = stream_lease.stream

            if new_frame:
                stream.frames[frame_id] = Frame()
                frame = stream.frames[frame_id]
                graph = self.pipeline_graph
            elif frame_id in stream.frames:
                frame = stream.frames[frame_id]
                graph = self.pipeline_graph.iterate_after(frame.paused_pe_name)
            else:
                self.logger.warning(
                    f"Process frame <{stream_id}:{frame_id}> frame not paused")

        if frame:
            frame.swag.update(frame_data_in)  # SWAG: Stuff We All Get 😅
        return graph, stream

# TODO: Refactor metrics into "utilities/metrics.py:class Metrics" ?

    def _process_metrics_initialize(self, frame):
        metrics = frame.metrics
        if metrics == {}:
            metrics["pipeline_elements"] = {}
            metrics["time_pipeline_start"] = time.time()
        return metrics

    def _process_metrics_capture(self, metrics, element_name, start_time):
        time_element = time.time() - start_time
        metrics["pipeline_elements"][f"time_{element_name}"] = time_element

        time_pipeline = time.time() - metrics["time_pipeline_start"]
        metrics["time_pipeline"] = time_pipeline  # Total time, so far !

    def _process_map_in(self, header, element, element_name, swag):
        map_in_names = {}
        if element_name in self.definition.map_in_nodes:
            map_in_elements = self.definition.map_in_nodes[element_name]
            for in_element, in_map in map_in_elements.items():
                from_name, to_name = next(iter(in_map.items()))
                map_in_names[to_name] = f"{element_name}.{to_name}"

        inputs = {}
        input_names = [input["name"] for input in element.definition.input]

        for input_name in input_names:
            try:
                if input_name in map_in_names:
                    inputs[input_name] = swag[map_in_names[input_name]]
                else:
                    inputs[input_name] = swag[input_name]
            except KeyError as key_error:
                self._error_pipeline(header,
                    f'Function parameter "{input_name}" not found')
        return inputs

    def _process_map_out(self, element_name, frame_data_out):
        if element_name in self.definition.map_out_nodes:
            map_out_node = self.definition.map_out_nodes[element_name]
            for out_element, out_map in map_out_node.items():
                from_name, to_name = next(iter(out_map.items()))
                to_name = f"{out_element}.{to_name}"
                frame_data_out[to_name] = frame_data_out.pop(from_name)

# FIX: _create_frame_generator(): StreamEvent.ERROR -->
#          self.destroy_stream(get_stream_id())  # immediately !

# TODO: Check local cases "stream_state.RUN | STOP | ERROR" ...
# TODO: - _create_frame_generator()
# TODO: - create_stream()
# TODO: - _process_frame_common()
# TODO: - destroy_stream()

# TODO: Check remote cases "stream_state.RUN | STOP | ERROR" ...
# TODO: - _create_frame_generator()
# TODO: - create_stream()
# TODO: - _process_frame_common()
# TODO: - destroy_stream()

    def _process_stream_event(self, element_name, stream_event, diagnostic,
        in_destroy_stream=False):

        def get_diagnostic(diagnostic):
            event_name = StreamEventName[stream_event]
            if "diagnostic" in diagnostic:
                diagnostic = diagnostic["diagnostic"]
            else:
                diagnostic = "No diagnostic provided"
            return f"{element_name.upper()}: {event_name} "  \
                   f"stream {self.my_id()} {diagnostic}"

        def get_stream_id():
            stream, _ = self.get_stream()
            return stream.stream_id

        stream_state = StreamState.RUN

        if stream_event == StreamEvent.STOP:
            stream_state = StreamState.STOP
            self.logger.debug(get_diagnostic(diagnostic))
            if not in_destroy_stream:  # avoid destroy_stream() recursion
                self._post_message(    # gracefully after frames processed
                    ActorTopic.IN, "destroy_stream", [get_stream_id()])

        elif stream_event == StreamEvent.ERROR:
            stream_state = StreamState.ERROR
            self.logger.error(get_diagnostic(diagnostic))
            if not in_destroy_stream:  # avoid destroy_stream() recursion
                self.destroy_stream(get_stream_id(), use_thread_local=False)

        return stream_state

    def set_parameter(self, stream_id, name, value):
        if stream_id in self.stream_leases:
            stream_lease = self.stream_leases[stream_id]
            parameters = stream_lease.stream.parameters
            parameters[name] = value

class PipelineRemoteAbsent(PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)
        self.share["lifecycle"] = "absent"

    def create_stream(self, stream_id, parameters=None, grace_time=_GRACE_TIME,
        queue_response=None, topic_response=None):
        self.log_error("create_stream")
        return False

    def destroy_stream(self, stream_id):
        self.log_error("destroy_stream")

    @classmethod
    def is_local(cls):
        return False

    def log_error(self, function_name):
        self.logger.error(f"PipelineElement.{function_name}(): "
                          f"{self.definition.name}: invoked when "
                           "remote Pipeline Actor hasn't been discovered")

    def process_frame(self, stream, **kwargs) -> Tuple[StreamEvent, dict]:
        self.log_error("process_frame")
        return False

class PipelineRemoteFound(PipelineElement):
    def __init__(self, context):
        context.get_implementation("PipelineElement").__init__(self, context)
        self.share["lifecycle"] = "ready"

    def create_stream(self, stream_id, parameters=None, grace_time=_GRACE_TIME,
        queue_response=None, topic_response=None):
        pass

    def destroy_stream(self, stream_id):
        pass

    @classmethod
    def is_local(cls):
        return False

# --------------------------------------------------------------------------- #

try:
    PIPELINE_DEFINITION_VERSION = 0
    PIPELINE_DEFINITION_SCHEMA = avro.schema.parse("""
{
  "namespace": "aiko_services",
  "name":      "pipeline_definition",
  "type":      "record",
  "fields": [
    { "name": "comment", "type": "string" },
    { "name": "version", "type": "int", "default": 0 },
    { "name": "name",    "type": "string" },
    { "name": "runtime", "type": {
        "name": "type",
        "type": "enum",
        "symbols": ["go", "python"]
      }
    },

    { "name": "graph", "type": {
        "type": "array", "items": "string"
      }
    },

    { "name": "parameters", "type": {
        "type": "map", "values": ["boolean", "int", "null", "string"]
      }
    },

    { "name": "elements", "type": [
        { "type": "array",
          "items": {
            "name": "element",
            "type": "record",
            "fields": [
              { "name": "comment", "type": "string" },
              { "name": "name",    "type": "string" },
              { "name": "input",   "type": {
                  "type": "array", "items": {
                    "name": "input",
                    "type": "record",
                    "fields": [
                      { "name": "type", "type": "string" },
                      { "name": "name", "type": "string" }
                    ]
                  }
                }
              },
              { "name": "output", "type": {
                  "type": "array", "items": {
                    "name": "output",
                    "type": "record",
                    "fields": [
                      { "name": "type", "type": "string" },
                      { "name": "name", "type": "string" }
                    ]
                  }
                }
              },
              { "name": "deploy", "type": {
                  "name": "deploy_fields",
                  "type": "record",
                  "fields": [
                    { "name": "local", "type": [
                        "null",
                        { "name": "local_fields",
                          "type": "record",
                          "fields": [
                            { "name": "class_name", "type": "string" },
                            { "name": "module", "type": "string" }
                          ]
                        }
                      ]
                    },
                    { "name": "remote", "type": [
                        "null",
                        { "name": "remote_fields",
                          "type": "record",
                          "fields": [
                            { "name": "module", "type": "string" },
                            { "name": "service_filter", "type": [
                                { "name": "service_filter",
                                  "type": "record",
                                  "fields": [
                                    { "name": "topic_path",
                                        "type": "string", "default": "*" },
                                    { "name": "name",
                                        "type": "string", "default": "*" },
                                    { "name": "owner",
                                        "type": "string", "default": "*" },
                                    { "name": "protocol",
                                        "type": "string", "default": "*" },
                                    { "name": "transport",
                                        "type": "string", "default": "*" },
                                    { "name": "tags",
                                        "type": "string", "default": "*" }
                                  ]
                                }
                              ]
                            }
                          ]
                        }
                      ]
                    }
                  ]
                }
              }
            ]
          }
        }
      ]
    }
  ]
}
    """)
except avro.errors.SchemaParseException as schema_parse_exception:
    PipelineImpl._exit(
        "Error: Parsing aiko_services/pipeline.py: PIPELINE_DEFINITION_SCHEMA",
        schema_parse_exception)

# --------------------------------------------------------------------------- #

@click.group()

def main():
    """Create and destroy Pipelines"""
    pass

@main.command(help="Create Pipeline defined by PipelineDefinition pathname")
@click.argument("definition_pathname", nargs=1, type=str)
@click.option("--name", "-n", type=str,
    default=None, required=False,
    help="Pipeline Actor name")
@click.option("--stream_id", "-s", type=str,
    default=None, required=False,
    help="Create Stream with identifier")
@click.option("--stream_parameters", "-sp", type=click.Tuple((str, str)),
    default=None, multiple=True, required=False,
    help="Define Stream parameters")
@click.option("--frame_id", "-fi", type=int,
    default=0, required=False,
    help="Process Frame with identifier")
@click.option("--frame_data", "-fd", type=str,
    default=None, required=False,
    help="Process Frame with data")
@click.option("--grace_time", "-gt", type=int,
    default=_GRACE_TIME, required=False,
    help="Stream receive frame time-out duration")
@click.option("--show_response", "-sr", is_flag=True,
    help="Show pipeline output response (output)")
@click.option("--log_level", "-ll", type=str,
    default="INFO", required=False,
    help="error, warning, info, debug")
@click.option("--log_mqtt", "-lm", type=str,
    default="all", required=False,
    help="all, false (console), true (mqtt)")

def create(definition_pathname, name, stream_id, stream_parameters,
    frame_id, frame_data, grace_time, show_response, log_level, log_mqtt):

    os.environ["AIKO_LOG_LEVEL"] = log_level.upper()
    os.environ["AIKO_LOG_MQTT"] = log_mqtt

    if not os.path.exists(definition_pathname):
        raise SystemExit(
            f"Error: PipelineDefinition not found: {definition_pathname}")

    pipeline_definition = PipelineImpl.parse_pipeline_definition(
        definition_pathname)

    def pipeline_response_handler(queue_pipeline_response):
        while True:
            response = queue_pipeline_response.get()
            id = f'<{response[0]["stream_id"]}:{response[0]["frame_id"]}>'
            frame_data = response[1]
            _LOGGER.info(f"Output: {id} {frame_data}")

    queue_pipeline_response = None
    if show_response:
        queue_pipeline_response = queue.Queue()
        Thread(target=pipeline_response_handler,
            args=(queue_pipeline_response,), daemon=True).start()

    pipeline = PipelineImpl.create_pipeline(
        definition_pathname, pipeline_definition,
        name, stream_id, stream_parameters, frame_id, frame_data, grace_time,
        queue_response=queue_pipeline_response)

    pipeline.run(mqtt_connection_required=False)

@main.command(help="Destroy Pipeline")
@click.argument("name", nargs=1, type=str, required=True)

def destroy(name):
# TODO: D.R.Y: Refactor based on "storage.py" and put this in the right place !
    def actor_discovery_handler(command, service_details):
        if command == "add":
            event.remove_timer_handler(waiting_timer)
            topic_path = f"{service_details[0]}/in"
            actor = get_actor_mqtt(topic_path, Pipeline)
            actor.stop()
            print(f'Destroyed Pipeline "{name}"')
            aiko.process.terminate()

    def waiting_timer():
        event.remove_timer_handler(waiting_timer)
        print(f'Waiting to discover Pipeline "{name}"')

    actor_discovery = ActorDiscovery(aiko.process)
    service_filter = ServiceFilter("*", name, "*", "*", "*", "*")
    actor_discovery.add_handler(actor_discovery_handler, service_filter)
    event.add_timer_handler(waiting_timer, 0.5)
    aiko.process.run()

if __name__ == "__main__":
    main()

# --------------------------------------------------------------------------- #



# These class definitions exist to workaround a short-coming in the
# PipelineElement Definition for "deploy":"remote" PipelineElements

from aiko_services.examples.pipeline.elements import PE_Add

# pipeline_small_*.json
class PE_A0(PE_Add): pass
class PE_B0(PE_Add): pass
class PE_C0(PE_Add): pass

# pipeline_large_*.json
class PE_000(PE_Add): pass
class PE_010(PE_Add): pass
class PE_020(PE_Add): pass
class PE_030(PE_Add): pass
class PE_040(PE_Add): pass
class PE_050(PE_Add): pass
class PE_060(PE_Add): pass
class PE_070(PE_Add): pass
class PE_080(PE_Add): pass
class PE_090(PE_Add): pass
